import functools
import inspect
import json
import os
import re
import typing
import uuid
from datetime import (
    date,
    datetime,
    time,
    timedelta
)
from time import (
    mktime,
    struct_time
)
from uuid import UUID

pascal_case_pattern = re.compile(r'(?<!^)(?=[A-Z])')

email_pattern = re.compile(r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)")
ip_address_pattern = re.compile(
    r"(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\."
    r"(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\."
    r"(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\."
    r"(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$"
)


def empty_object_id():
    return uuid.UUID('00000000-0000-0000-0000-000000000000')


def new_object_id():
    return uuid.uuid4()


def current_time():
    return datetime.now()


def new_nickname():
    return f'random_nickname_{uuid.uuid4()}'


def new_password():
    return f'random_password_{uuid.uuid4()}'


def mail_validator(email):
    return email_pattern.match(email) is not None


def ip_validator(ip):
    return ip_address_pattern.match(ip) is not None


def to_snake_string(string: str):
    return pascal_case_pattern.sub('_', string).lower()


def to_pascal_string(string: str):
    return ''.join(word.title() for word in string.split('_'))


def to_upper_string(string: str):
    return string.upper()


def to_lower_string(string: str):
    return string.lower()


def to_title(string: str):
    return string.title()


def to_enumerable(value, string=False):
    if value is None:
        return None

    from joatmon.serializable import Serializable
    from joatmon.orm.enum import Enum

    if isinstance(value, dict):
        ret = {k: to_enumerable(v, string) for k, v in value.items()}
    elif isinstance(value, list):
        ret = [to_enumerable(v, string) for v in value]
    elif isinstance(value, tuple):
        ret = tuple([to_enumerable(v, string) for v in value])
    elif isinstance(value, Serializable):
        ret = to_enumerable({k: v for k, v in value.__dict__.items()}, string)
    elif isinstance(value, Enum):
        ret = str(value).lower()
    else:
        ret = value
        if string:
            ret = str(ret)
    return ret


def to_case(case, value, key=None, convert_value=False):
    from joatmon.serializable import Serializable

    enumerable = to_enumerable(value)

    if case == 'snake':
        new_key = to_snake_string(key) if key is not None else None
    elif case == 'pascal':
        new_key = to_pascal_string(key) if key is not None else None
    elif case == 'upper':
        new_key = to_upper_string(key) if key is not None else None
    elif case == 'lower':
        new_key = to_lower_string(key) if key is not None else None
    else:
        raise ValueError(f'{case} is not supported case.')

    if isinstance(enumerable, dict):
        ret = new_key, {k: v for k, v in [to_case(case, v, k) for k, v in enumerable.items()]}
    elif isinstance(enumerable, list):
        ret = new_key, [to_case(case, v) for v in enumerable]
    elif isinstance(enumerable, tuple):
        ret = new_key, tuple([to_case(case, v) for v in enumerable])
    elif isinstance(value, Serializable):
        ret = new_key, to_case(case, value.__dict__)
    else:
        if convert_value:
            value = to_snake_string(value)
        ret = new_key, value

    if new_key is None:
        return ret[1]
    return ret


def get_function_args(func, *args):
    if get_class_that_defined_method(func) is None:
        return args
    else:
        # need to check if the function is static / class method
        # if static no need to remove first args
        return args[1:]


def get_function_kwargs(func, **kwargs):
    return kwargs


def to_hash(func, *args, **kwargs):
    args_str = ", ".join([f"{arg}" for arg in get_function_args(func, *args)])
    kwargs_str = ", ".join([f"{k}={v}" for k, v in get_function_kwargs(func, **kwargs).items()])
    arg_kwarg_str = ', '.join(filter(lambda x: x != '', [args_str, kwargs_str]))
    return f'{func.__module__}.{func.__qualname__}({arg_kwarg_str})'


def get_converter(kind: type):
    def _datetime_converter(value: object) -> typing.Optional[datetime]:
        if value is None:
            return value

        if isinstance(value, datetime):
            return value

        if isinstance(value, struct_time):
            return datetime.fromtimestamp(mktime(value))

        if isinstance(value, str):
            return datetime.fromisoformat(value)

        raise ValueError(f'cannot convert {value} with type {type(value)} to datetime')

    def _integer_converter(value: object) -> typing.Optional[int]:
        if value is None:
            return value

        if isinstance(value, float):
            return int(value)

        if isinstance(value, int):
            return int(value)

        if isinstance(value, str):
            return int(value)

        raise ValueError(f'cannot convert {value} with type {type(value)} to integer')

    def _float_converter(value: object) -> typing.Optional[float]:
        if value is None:
            return value

        if isinstance(value, float):
            return value

        if isinstance(value, int):
            return float(value)

        if isinstance(value, str):
            return float(value)

        raise ValueError(f'cannot convert {value} with type {type(value)} to float')

    def _string_converter(value: object) -> typing.Optional[str]:
        if value is None:
            return value

        if isinstance(value, str):
            return value

        # need to convert bytes as well

        raise ValueError(f'cannot convert {value} with type {type(value)} to string')

    def _byte_converter(value: object) -> typing.Optional[bytes]:
        if value is None:
            return value

        if isinstance(value, bytes):
            return value

        # need to convert string and int list as well

        raise ValueError(f'cannot convert {value} with type {type(value)} to bytes')

    def _boolean_converter(value: object) -> typing.Optional[bool]:
        if value is None:
            return value

        if isinstance(value, bool):
            return value

        # need to convert float, int and str as well

        raise ValueError(f'cannot convert {value} with type {type(value)} to boolean')

    def _uuid_converter(value: object) -> typing.Optional[uuid.UUID]:
        if value is None:
            return value

        if isinstance(value, uuid.UUID):
            return value

        if isinstance(value, str):
            return uuid.UUID(value)

        raise ValueError(f'cannot convert {value} with type {type(value)} to uuid')

    def _dictionary_converter(value: object) -> typing.Optional[dict]:
        if value is None:
            return value

        if isinstance(value, dict):
            return to_enumerable(value)

        if hasattr(value, 'dict'):
            return to_enumerable(value)

        raise ValueError(f'cannot convert {value} with type {type(value)} to dictionary')

    def _list_converter(value: object) -> typing.Optional[list]:
        if value is None:
            return value

        # if isinstance(value, str):
        #     value = json.loads(value)

        if isinstance(value, (list, tuple, set)):
            ret = []
            for v in value:
                ret.append(to_enumerable(v))
            return ret

        raise ValueError(f'cannot convert {value} with type {type(value)} to list')

    def _tuple_converter(value: object) -> typing.Optional[tuple]:
        if value is None:
            return value

        # if isinstance(value, str):
        #     value = json.loads(value)

        if isinstance(value, (list, tuple, set)):
            ret = []
            for v in value:
                ret.append(to_enumerable(v))
            ret = tuple(ret)
            return ret

        raise ValueError(f'cannot convert {value} with type {type(value)} to tuple')

    def _set_converter(value: object) -> typing.Optional[set]:
        if value is None:
            return value

        # if isinstance(value, str):
        #     value = json.loads(value)

        if isinstance(value, (list, tuple, set)):
            ret = []
            for v in value:
                ret.append(to_enumerable(v))
            ret = set(ret)
            return ret

        raise ValueError(f'cannot convert {value} with type {type(value)} to tuple')

    def _object_converter(value: object) -> typing.Optional[object]:
        if value is None:
            return value

        if isinstance(value, dict):
            return _dictionary_converter(value)

        if isinstance(value, list):
            return _list_converter(value)

        if isinstance(value, tuple):
            return _tuple_converter(value)

        if isinstance(value, str):
            return _string_converter(value)

        if hasattr(value, 'dict'):
            return value.dict

        return value

    converters = {
        datetime: _datetime_converter,
        int: _integer_converter,
        float: _float_converter,
        str: _string_converter,
        bytes: _byte_converter,
        bool: _boolean_converter,
        UUID: _uuid_converter,
        dict: _dictionary_converter,
        list: _list_converter,
        tuple: _tuple_converter,
        set: _set_converter,
        object: _object_converter
    }

    return converters.get(kind, _object_converter)


def to_list(items):
    ret = []
    for item in iter(items):
        ret.append(item)
    return ret


async def to_list_async(items):
    ret = []
    async for item in aiter(items):
        ret.append(item)
    return ret


def first(items):
    try:
        iterable = iter(items)
        first_item = next(iterable)
        return first_item
    except StopIteration:
        return None


async def first_async(items):
    try:
        iterable = aiter(items)
        first_item = await anext(iterable)
        return first_item
    except StopAsyncIteration:
        return None


def single(items):
    try:
        iterable = iter(items)
        first_item = next(iterable)
    except StopIteration:
        return None

    try:
        next(iterable)
        return None
    except StopIteration:
        return first_item


async def single_async(items):
    try:
        iterable = aiter(items)
        first_item = await anext(iterable)
    except StopAsyncIteration:
        return None

    try:
        await anext(iterable)
        return None
    except StopAsyncIteration:
        return first_item


def pretty_printer(headers, m=None):
    def get_max_size(idx):
        max_size = m or os.get_terminal_size().columns
        weight = headers[idx][1]
        # if idx == len(headers) - 1:
        #     return max_size - (int(max_size * (weight / sum(x[1] for x in headers))) * len(headers) - 1)
        # else:
        #     return int(max_size * (weight / sum(x[1] for x in headers)))
        return int(max_size * (weight / sum(x[1] for x in headers)))

    def left_padding(idx, value):
        max_size = get_max_size(idx)
        return (max_size - len(value)) // 2

    def right_padding(idx, value):
        max_size = get_max_size(idx)
        return max_size - (((max_size - len(value)) // 2) + len(value))

    def prettify_header(idx, value):
        if len(value) > get_max_size(idx):
            return value[:get_max_size(idx)]

        return ' ' * left_padding(idx, value) + value + ' ' * right_padding(idx, value)

    def prettify_value(idx, value):
        if len(value) > get_max_size(idx):
            return value[:get_max_size(idx)]

        return ' ' * left_padding(idx, value) + value + ' ' * right_padding(idx, value)

    def pretty_print(values):
        return ' '.join([prettify_value(idx, value) for idx, value in enumerate(values)])

    return ' '.join([prettify_header(idx, header) for idx, (header, max_size) in enumerate(headers)]), pretty_print


def convert_size(size_bytes):
    import math
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])


def get_class_that_defined_method(meth):
    if isinstance(meth, functools.partial):
        return get_class_that_defined_method(meth.func)
    if inspect.ismethod(meth) or (inspect.isbuiltin(meth) and getattr(meth, '__self__', None) is not None and getattr(meth.__self__, '__class__', None)):
        for cls in inspect.getmro(meth.__self__.__class__):
            if meth.__name__ in cls.__dict__:
                return cls
        meth = getattr(meth, '__func__', meth)  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth), meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0], None)
        if isinstance(cls, type):
            return cls
    return getattr(meth, '__objclass__', None)  # handle special descriptor objects


def get_module_functions(module):
    return inspect.getmembers(module, inspect.isfunction)


def get_module_classes(module):
    return inspect.getmembers(module, inspect.isclass)


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime, date, time)):
            return obj.isoformat()
        if isinstance(obj, timedelta):
            return (datetime.min + obj).time().isoformat()
        if isinstance(obj, uuid.UUID):
            return str(obj)
        if callable(obj):
            return str(obj)
        return str(obj)


class JSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        super(JSONDecoder, self).__init__(object_hook=self.object_hook, *args, **kwargs)

    @staticmethod
    def object_hook(d):
        return d
        # ret = {}
        # for key, value in d.items():
        #     print('__type__' in value, key)
        #     if '__type__' not in value:
        #         ret[key] = value
        #         continue
        #
        #     t = value.pop('__type__')
        #     try:
        #         if t == 'datetime':
        #             ret[key] = datetime(**value)
        #     except:
        #         value['__type___'] = t
        #         ret[key] = value
        # return ret



from joatmon.nn.core import *
from joatmon.nn.functional import *
from joatmon.nn.init import *
from joatmon.nn.layer import *
from joatmon.nn.loss import *
from joatmon.nn.optimizer import *
from joatmon.nn.scheduler import *

# from joatmon.ai.nn.utility import *

Tensor.chunk = chunk
Tensor.view = view
Tensor.index_select = index_select

Tensor.zero = zero
Tensor.one = one
Tensor.fill = fill
Tensor.squeeze = squeeze
Tensor.expand_dim = expand_dim
Tensor.transpose = transpose
Tensor.absolute = absolute
Tensor.around = around
Tensor.floor = floor
Tensor.ceil = ceil
Tensor.clip = clip
Tensor.negative = negative
Tensor.summation = summation
Tensor.mean = mean
Tensor.std = std
Tensor.var = var
Tensor.var = var
Tensor.add = add
Tensor.sub = sub
Tensor.mul = mul
Tensor.div = div
Tensor.power = power
Tensor.clone = clone
Tensor.detach = detach

Tensor.from_array = from_array
Tensor.to_array = to_array

Tensor.half = half
Tensor.single = single
Tensor.double = double
Tensor.cpu = cpu
Tensor.gpu = gpu



from joatmon.nn import functional as f
from joatmon.nn.core import Module

__all__ = ['Flatten']


class Flatten(Module):
    def __init__(self):
        super(Flatten, self).__init__()

    def forward(self, inp):
        return f.view(
            inp=inp
        )



import asyncio
import copy
import logging
import ssl
from collections import deque
from functools import wraps
from urllib.parse import (
    urlparse,
    urlunparse
)

import websockets
from websockets.exceptions import InvalidHandshake
from websockets.uri import InvalidURI

from .adapters import (
    StreamReaderAdapter,
    StreamWriterAdapter,
    WebSocketsReader,
    WebSocketsWriter
)
from .mqtt.constants import (
    QOS_0,
    QOS_1,
    QOS_2
)
from .mqtt.handler import (
    ClientProtocolHandler,
    ProtocolHandlerException
)
from .mqtt.packet import CONNECTION_ACCEPTED
from .session import Session
from .utils import not_in_dict_or_none

_defaults = {
    'keep_alive': 10,
    'ping_delay': 1,
    'default_qos': 0,
    'default_retain': False,
    'auto_reconnect': True,
    'reconnect_max_interval': 10,
    'reconnect_retries': 2,
}


class ClientException(Exception):
    pass


class ConnectException(ClientException):
    pass


base_logger = logging.getLogger(__name__)


def mqtt_connected(func):
    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        if not self._connected_state.is_set():
            base_logger.warning("Client not connected, waiting for it")
            _, pending = await asyncio.wait([self._connected_state.wait(), self._no_more_connections.wait()], return_when=asyncio.FIRST_COMPLETED)
            for t in pending:
                t.cancel()
            if self._no_more_connections.is_set():
                raise ClientException("Will not reconnect")
        return await func(self, *args, **kwargs)

    return wrapper


class MQTTClient:
    def __init__(self, client_id=None, config=None, loop=None):
        self.logger = logging.getLogger(__name__)
        self.config = copy.deepcopy(_defaults)
        if config is not None:
            self.config.update(config)
        if client_id is not None:
            self.client_id = client_id
        else:
            from .utils import gen_client_id
            self.client_id = gen_client_id()
            self.logger.debug("Using generated client ID : %s" % self.client_id)

        if loop is not None:
            self._loop = loop
        else:
            self._loop = asyncio.get_event_loop()
        self.session = None
        self._handler = None
        self._disconnect_task = None
        self._connected_state = asyncio.Event(loop=self._loop)
        self._no_more_connections = asyncio.Event(loop=self._loop)
        self.extra_headers = {}

        self.client_tasks = deque()

    async def connect(self, uri=None, cleansession=None, cafile=None, capath=None, cadata=None, extra_headers={}):
        self.session = self._initsession(uri, cleansession, cafile, capath, cadata)
        self.extra_headers = extra_headers
        self.logger.debug("Connect to: %s" % uri)

        try:
            return await self._do_connect()
        except BaseException as be:
            self.logger.warning("Connection failed: %r" % be)
            auto_reconnect = self.config.get('auto_reconnect', False)
            if not auto_reconnect:
                raise
            else:
                return await self.reconnect()

    async def disconnect(self):
        await self.cancel_tasks()
        if self.session.transitions.is_connected():
            if not self._disconnect_task.done():
                self._disconnect_task.cancel()
            await self._handler.mqtt_disconnect()
            self._connected_state.clear()
            await self._handler.stop()
            self.session.transitions.disconnect()
        else:
            self.logger.warning("Client session is not currently connected, ignoring call")

    async def cancel_tasks(self):
        try:
            while self.client_tasks:
                task = self.client_tasks.pop()
                task.cancel()
        except IndexError as err:
            pass

    async def reconnect(self, cleansession=None):
        if self.session.transitions.is_connected():
            self.logger.warning("Client already connected")
            return CONNECTION_ACCEPTED

        if cleansession:
            self.session.clean_session = cleansession
        self.logger.debug("Reconnecting with session parameters: %s" % self.session)
        reconnect_max_interval = self.config.get('reconnect_max_interval', 10)
        reconnect_retries = self.config.get('reconnect_retries', 5)
        nb_attempt = 1
        await asyncio.sleep(1, loop=self._loop)
        while True:
            try:
                self.logger.debug("Reconnect attempt %d ..." % nb_attempt)
                return await self._do_connect()
            except BaseException as e:
                self.logger.warning("Reconnection attempt failed: %r" % e)
                if 0 <= reconnect_retries < nb_attempt:
                    self.logger.error("Maximum number of connection attempts reached. Reconnection aborted")
                    raise ConnectException("Too many connection attempts failed")
                exp = 2 ** nb_attempt
                delay = exp if exp < reconnect_max_interval else reconnect_max_interval
                self.logger.debug("Waiting %d second before next attempt" % delay)
                await asyncio.sleep(delay, loop=self._loop)
                nb_attempt += 1

    async def _do_connect(self):
        return_code = await self._connect_coro()
        self._disconnect_task = asyncio.ensure_future(self.handle_connection_close(), loop=self._loop)
        return return_code

    @mqtt_connected
    async def ping(self):
        if self.session.transitions.is_connected():
            await self._handler.mqtt_ping()
        else:
            self.logger.warning("MQTT PING request incompatible with current session state '%s'" % self.session.transitions.state)

    @mqtt_connected
    async def publish(self, topic, message, qos=None, retain=None, ack_timeout=None):
        def get_retain_and_qos():
            if qos:
                assert qos in (QOS_0, QOS_1, QOS_2)
                _qos = qos
            else:
                _qos = self.config['default_qos']
                try:
                    _qos = self.config['topics'][topic]['qos']
                except KeyError:
                    pass
            if retain:
                _retain = retain
            else:
                _retain = self.config['default_retain']
                try:
                    _retain = self.config['topics'][topic]['retain']
                except KeyError:
                    pass
            return _qos, _retain

        (app_qos, app_retain) = get_retain_and_qos()
        return await self._handler.mqtt_publish(topic, message, app_qos, app_retain, ack_timeout)

    @mqtt_connected
    async def subscribe(self, topics):
        return await self._handler.mqtt_subscribe(topics, self.session.next_packet_id)

    @mqtt_connected
    async def unsubscribe(self, topics):
        await self._handler.mqtt_unsubscribe(topics, self.session.next_packet_id)

    async def deliver_message(self, timeout=None):
        deliver_task = asyncio.ensure_future(self._handler.mqtt_deliver_next_message(), loop=self._loop)
        self.client_tasks.append(deliver_task)
        self.logger.debug("Waiting message delivery")
        done, pending = await asyncio.wait([deliver_task], loop=self._loop, return_when=asyncio.FIRST_EXCEPTION, timeout=timeout)
        if self.client_tasks:
            self.client_tasks.pop()
        if deliver_task in done:
            if deliver_task.exception() is not None:
                raise deliver_task.exception()
            return deliver_task.result()
        else:
            deliver_task.cancel()
            raise asyncio.TimeoutError

    async def _connect_coro(self):
        kwargs = dict()

        uri_attributes = urlparse(self.session.broker_uri)
        scheme = uri_attributes.scheme
        secure = True if scheme in ('mqtts', 'wss') else False
        self.session.username = self.session.username if self.session.username else uri_attributes.username
        self.session.password = self.session.password if self.session.password else uri_attributes.password
        self.session.remote_address = uri_attributes.hostname
        self.session.remote_port = uri_attributes.port
        if scheme in ('mqtt', 'mqtts') and not self.session.remote_port:
            self.session.remote_port = 8883 if scheme == 'mqtts' else 1883
        if scheme in ('ws', 'wss') and not self.session.remote_port:
            self.session.remote_port = 443 if scheme == 'wss' else 80
        if scheme in ('ws', 'wss'):
            uri = (scheme, self.session.remote_address + ":" + str(self.session.remote_port), uri_attributes[2], uri_attributes[3], uri_attributes[4], uri_attributes[5])
            self.session.broker_uri = urlunparse(uri)
        self._handler = ClientProtocolHandler(loop=self._loop)

        if secure:
            sc = ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cafile=self.session.cafile, capath=self.session.capath, cadata=self.session.cadata)
            if 'certfile' in self.config and 'keyfile' in self.config:
                sc.load_cert_chain(self.config['certfile'], self.config['keyfile'])
            if 'check_hostname' in self.config and isinstance(self.config['check_hostname'], bool):
                sc.check_hostname = self.config['check_hostname']
            kwargs['ssl'] = sc

        try:
            reader = None
            writer = None
            self._connected_state.clear()

            if scheme in ('mqtt', 'mqtts'):
                conn_reader, conn_writer = await asyncio.open_connection(self.session.remote_address, self.session.remote_port, loop=self._loop, **kwargs)
                reader = StreamReaderAdapter(conn_reader)
                writer = StreamWriterAdapter(conn_writer)
            elif scheme in ('ws', 'wss'):
                websocket = await websockets.connect(self.session.broker_uri, subprotocols=['mqtt'], loop=self._loop, extra_headers=self.extra_headers, **kwargs)
                reader = WebSocketsReader(websocket)
                writer = WebSocketsWriter(websocket)

            self._handler.attach(self.session, reader, writer)
            return_code = await self._handler.mqtt_connect()
            if return_code is not CONNECTION_ACCEPTED:
                self.session.transitions.disconnect()
                self.logger.warning("Connection rejected with code '%s'" % return_code)
                exc = ConnectException("Connection rejected by broker")
                exc.return_code = return_code
                raise exc
            else:
                await self._handler.start()
                self.session.transitions.connect()
                self._connected_state.set()
                self.logger.debug("connected to %s:%s" % (self.session.remote_address, self.session.remote_port))
            return return_code
        except InvalidURI as iuri:
            self.logger.warning("connection failed: invalid URI '%s'" % self.session.broker_uri)
            self.session.transitions.disconnect()
            raise ConnectException("connection failed: invalid URI '%s'" % self.session.broker_uri, iuri)
        except InvalidHandshake as ihs:
            self.logger.warning("connection failed: invalid websocket handshake")
            self.session.transitions.disconnect()
            raise ConnectException("connection failed: invalid websocket handshake", ihs)
        except (ProtocolHandlerException, ConnectionError, OSError) as e:
            self.logger.warning("MQTT connection failed: %r" % e)
            self.session.transitions.disconnect()
            raise ConnectException(e)

    async def handle_connection_close(self):

        def cancel_tasks():
            self._no_more_connections.set()
            while self.client_tasks:
                task = self.client_tasks.popleft()
                if not task.done():
                    task.cancel()

        self.logger.debug("Watch broker disconnection")
        await self._handler.wait_disconnect()
        self.logger.warning("Disconnected from broker")

        self._connected_state.clear()

        self._handler.detach()
        self.session.transitions.disconnect()

        if self.config.get('auto_reconnect', False):
            self.logger.debug("Auto-reconnecting")
            try:
                await self.reconnect()
            except ConnectException:
                cancel_tasks()
        else:
            cancel_tasks()

    def _initsession(self, uri=None, cleansession=None, cafile=None, capath=None, cadata=None) -> Session:
        broker_conf = self.config.get('broker', dict()).copy()
        if uri:
            broker_conf['uri'] = uri
        if cafile:
            broker_conf['cafile'] = cafile
        elif 'cafile' not in broker_conf:
            broker_conf['cafile'] = None
        if capath:
            broker_conf['capath'] = capath
        elif 'capath' not in broker_conf:
            broker_conf['capath'] = None
        if cadata:
            broker_conf['cadata'] = cadata
        elif 'cadata' not in broker_conf:
            broker_conf['cadata'] = None

        if cleansession is not None:
            broker_conf['cleansession'] = cleansession

        for key in ['uri']:
            if not_in_dict_or_none(broker_conf, key):
                raise ClientException("Missing connection parameter '%s'" % key)

        s = Session()
        s.broker_uri = uri
        s.client_id = self.client_id
        s.cafile = broker_conf['cafile']
        s.capath = broker_conf['capath']
        s.cadata = broker_conf['cadata']
        if cleansession is not None:
            s.clean_session = cleansession
        else:
            s.clean_session = self.config.get('cleansession', True)
        s.keep_alive = self.config['keep_alive'] - self.config['ping_delay']
        if 'will' in self.config:
            s.will_flag = True
            s.will_retain = self.config['will']['retain']
            s.will_topic = self.config['will']['topic']
            s.will_message = self.config['will']['message']
            s.will_qos = self.config['will']['qos']
        else:
            s.will_flag = False
            s.will_retain = False
            s.will_topic = None
            s.will_message = None
        print(s.username, s.password, s.client_id)
        return s



import typing
import warnings
from collections import OrderedDict
from datetime import datetime
from uuid import UUID

import pymongo
from bson.binary import UuidRepresentation
from bson.codec_options import DEFAULT_CODEC_OPTIONS
from pymongo import (
    read_concern,
    write_concern
)

from joatmon.orm.constraint import UniqueConstraint
from joatmon.plugin.database.core import DatabasePlugin


class MongoDatabase(DatabasePlugin):
    DATABASES = set()
    CREATED_COLLECTIONS = set()
    UPDATED_COLLECTIONS = set()

    def __init__(self, uri, database):
        self.database_name = database
        self.client = pymongo.MongoClient(host=uri)
        self.database = self.client[database]

        self.session = None

    async def _check_collection(self, collection):
        return collection.__collection__ in list(self.database.list_collection_names())

    async def _create_collection(self, collection):
        def get_type(dtype: typing.Union[type, typing.List, typing.Tuple]):
            type_mapper = {
                datetime: ['date'],
                int: ['int'],
                float: ['double'],
                str: ['string'],
                bool: ['bool'],
                UUID: ['binData'],
                dict: ['object'],
                list: ['array'],
                tuple: ['array'],
                object: ['object']
            }

            if isinstance(dtype, (tuple, list)):
                return sum(list(map(lambda x: type_mapper.get(x, ['object']), dtype)), [])
            else:
                return type_mapper.get(dtype, ['object'])

        self.database.create_collection(
            collection.__collection__
        )

        vexpr = {
            '$jsonSchema': {
                'bsonType': 'object',
                'required': list(
                    map(lambda x: x[0], filter(lambda y: not y[1].nullable, collection.fields(collection).items()))
                ),
                'properties': {}
            }
        }
        for field_name, field in collection.fields(collection).items():
            # might want to rewrite this part again
            vexpr['$jsonSchema']['properties'][field_name] = {
                'bsonType': list(
                    set(get_type(field.dtype) if not field.nullable else get_type(field.dtype) + ['null'])
                ),
                'description': ''
            }

        cmd = OrderedDict(
            [  # indexes can be added here as well
                ('collMod', collection.__collection__),
                ('validator', vexpr),
                ('validationLevel', 'moderate')
            ]
        )

        self.database.command(cmd)

        index_names = set()
        for index_name, index in collection.constraints(collection).items():
            if ',' in index.field:
                index_fields = list(map(lambda x: x.strip(), index.field.split(',')))
            else:
                index_fields = [index.field]
            c = [(f'{k}', 1) for k in index_fields]
            if index_name in index_names:
                continue
            index_names.add(index_name)
            try:
                self.database[collection.__collection__].create_index(c, unique=isinstance(index, UniqueConstraint), name=index_name)
            except Exception as ex:
                print(str(ex))

    async def _ensure_collection(self, collection):
        if not await self._check_collection(collection):
            await self._create_collection(collection)

    async def _get_collection(self, collection):
        codec_options = DEFAULT_CODEC_OPTIONS.with_options(uuid_representation=UuidRepresentation.STANDARD)
        if self.session is None:
            return self.database.get_collection(collection, codec_options=codec_options)
        else:
            return self.session.client[self.database_name].get_collection(collection, codec_options=codec_options)

    async def insert(self, document, *docs):
        for doc in docs:
            d = dict(**doc)

            if document.__metaclass__.structured and document.__metaclass__.force:
                await self._ensure_collection(document.__metaclass__)
                d = document(**doc).validate()
            elif document.__metaclass__.structured:
                warnings.warn(f'document validation will be ignored')

            collection = await self._get_collection(document.__metaclass__.__collection__)
            collection.insert_one(d, session=self.session)

    async def read(self, document, query):
        if document.__metaclass__.structured and document.__metaclass__.force:
            await self._ensure_collection(document.__metaclass__)
        elif document.__metaclass__.structured:
            warnings.warn(f'document validation will be ignored')

        collection = await self._get_collection(document.__metaclass__.__collection__)
        result = collection.find(dict(**query), {'_id': 0}, session=self.session)

        for doc in result:
            yield document(**doc)

    async def update(self, document, query, update):
        if document.__metaclass__.structured and document.__metaclass__.force:
            await self._ensure_collection(document.__metaclass__)
        elif document.__metaclass__.structured:
            warnings.warn(f'document validation will be ignored')

        collection = await self._get_collection(document.__metaclass__.__collection__)
        collection.update_many(dict(**query), {'$set': dict(**update)}, session=self.session)

    async def delete(self, document, query):
        if document.__metaclass__.structured and document.__metaclass__.force:
            await self._ensure_collection(document.__metaclass__)
        elif document.__metaclass__.structured:
            warnings.warn(f'document validation will be ignored')

        collection = await self._get_collection(document.__metaclass__.__collection__)
        collection.delete_many(dict(**query), session=self.session)

    async def start(self):
        self.session = self.client.start_session()
        self.session.start_transaction(read_concern.ReadConcern('majority'), write_concern.WriteConcern('majority'))

    async def commit(self):
        self.session.commit_transaction()

    async def abort(self):
        self.session.abort_transaction()

    async def end(self):
        self.session.end_session()



import random

__all__ = ['TilesEnv']

from joatmon.game.core import CoreEnv


class TilesEnv(CoreEnv):
    def __init__(self, size):
        super().__init__()

        self.size = size

        self.tiles = []
        for y in range(self.size):
            row = []
            for x in range(self.size):
                if y * self.size + x + 1 == self.size ** 2:
                    row.append(0)
                else:
                    row.append(y * self.size + x + 1)
            self.tiles.append(row)

    def _find_empty(self):
        for y in range(self.size):
            for x in range(self.size):
                if self.tiles[y][x] == 0:
                    return y, x
        return None, None

    def _swap(self, pos1, pos2):
        temp = self.tiles[pos2[0]][pos2[1]]
        self.tiles[pos2[0]][pos2[1]] = self.tiles[pos1[0]][pos1[1]]
        self.tiles[pos1[0]][pos1[1]] = temp

    def close(self):
        pass

    def render(self, mode='human'):
        to_print = ''
        for y in range(self.size):
            line = ''
            for x in range(self.size):
                line += '{:2d} '.format(self.tiles[y][x]) \
                    if self.tiles[y][x] != 0 else '   '
            to_print += line + '\n'
        print(to_print)

    def reset(self):
        r = [i + 1 for i in range(self.size ** 2)]
        random.shuffle(r)

        for i, elem in enumerate(r):
            div, mod = divmod(i, self.size)
            if elem == self.size ** 2:
                self.tiles[div][mod] = 0
            else:
                self.tiles[div][mod] = elem

    def seed(self, seed=None):
        pass

    def step(self, action):
        empty_y, empty_x = self._find_empty()

        if empty_y is not None and empty_x is not None:
            if isinstance(action, int):
                if 0 <= action < 4:
                    if action == 0:
                        if empty_x + 1 < self.size:
                            self._swap([empty_y, empty_x], [empty_y, empty_x + 1])
                    elif action == 1:
                        if empty_y - 1 >= 0:
                            self._swap([empty_y, empty_x], [empty_y - 1, empty_x])
                    elif action == 2:
                        if empty_x - 1 >= 0:
                            self._swap([empty_y, empty_x], [empty_y, empty_x - 1])
                    elif action == 3:
                        if empty_y + 1 < self.size:
                            self._swap([empty_y, empty_x], [empty_y + 1, empty_x])



from __future__ import print_function

import json

import requests

from joatmon.assistant.task import BaseTask


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "weather",
            "description": "a function for user to learn current, historical, forecast weather data for given location",
            "parameters": {
                "type": "object",
                "properties": {
                    "mode": {
                        "type": "string",
                        "enum": ["current", "history", "forecast"]
                    },
                    "location": {
                        "type": "string",
                        "description": "location of the weather"
                    }
                },
                "required": ["mode", "location"]
            }
        }

    def run(self):
        key = json.loads(open('iva/iva.json', 'r').read())['config']['weather']['key']

        mode = self.kwargs.get('mode', None)
        location = self.kwargs.get('location', None) or self.api.input('what is the location')

        if mode == 'current':
            resp = requests.get('https://api.weatherapi.com/v1/current.json', params={'key': key, 'q': location})

            response = json.loads(resp.content.decode('utf-8'))
            current = response.get('current', {})
            temp_c = current.get('temp_c', None)
            condition = current.get('condition', {}).get('text', None)
            wind_kph = current.get('wind_kph', None)
            wind_degree = current.get('wind_degree', None)
            wind_dir = current.get('wind_dir', None)
            pressure_mb = current.get('pressure_mb', None)
            humidity = current.get('humidity', None)
            cloud = current.get('cloud', None)
            feelslike_c = current.get('feelslike_c', None)
            vis_km = current.get('vis_km', None)

            self.api.output(f'Current Temperature is: {temp_c} and it feels like {feelslike_c}, Current Humidity is: {humidity}')

        if not self.stop_event.is_set():
            self.stop_event.set()



from . import random
from .array import Array
from .functional import *

Array.astype = astype
Array.copy = copy
Array.repeat = repeat
Array.split = split
Array.tolist = tolist

Array.getitem = getitem
Array.take_along_axis = take_along_axis
Array.setitem = setitem
Array.put_along_axis = put_along_axis

Array.where = where
Array.indices = indices

Array.dim = dim
Array.size = size
Array.flatten = flatten
Array.reshape = reshape
Array.squeeze = squeeze
Array.expand_dims = expand_dims
Array.pad = pad
Array.transpose = transpose

Array.fill = fill

Array.abs = absolute
Array.negative = negative
Array.round = around
Array.floor = floor
Array.ceil = ceil
Array.sqrt = sqrt
Array.square = square
Array.clip = clip
Array.exp = exp
Array.tanh = tanh
Array.sum = sum
Array.mean = mean
Array.median = median
Array.var = var
Array.std = std
Array.prod = prod
Array.unique = unique
Array.argmax = argmax
Array.argmin = argmin
Array.amax = amax
Array.amin = amin

Array.add = add
Array.sub = sub
Array.truediv = truediv
Array.floordiv = floordiv
Array.mul = mul
Array.power = power
Array.lt = lt
Array.le = le
Array.gt = gt
Array.ge = ge
Array.eq = eq
Array.ne = ne



import redis

from joatmon.plugin.cache.core import CachePlugin


class RedisCache(CachePlugin):
    def __init__(self, host: str, port: int, password: str):
        self.host = host
        self.port = port
        self.password = password

        self.connection = redis.Redis(host=host, port=port, password=password)

    async def add(self, key, value, duration=None):
        self.connection.set(key, value, ex=duration)

    async def get(self, key):
        return self.connection.get(key)

    async def update(self, key, value):
        self.connection.set(key, value)

    async def remove(self, key):
        for k in self.connection.keys(key):
            self.connection.delete(k)



import queue
import threading

import speech_recognition as sr


class Microphone:
    def __init__(self):
        super(Microphone, self).__init__()

        self.r = sr.Recognizer()
        self.r.energy_threshold = 350
        self.r.pause_threshold = 0.8
        self.r.dynamic_energy_threshold = False
        self.r.timeout = 1

        self.audio_queue = queue.Queue()
        self.result_queue = queue.Queue()

        self.stop_event = threading.Event()

        self.listening_thread = threading.Thread(target=self.record_audio)

        # self.listening_thread.start()

    def record_audio(self):
        with sr.Microphone(sample_rate=16000) as source:
            # self.r.adjust_for_ambient_noise(source)
            while not self.stop_event.is_set():
                try:
                    audio = self.r.listen(source, timeout=1)
                    self.audio_queue.put_nowait(audio)
                except sr.exceptions.WaitTimeoutError:
                    ...

    def listen(self):
        return input()

        # return self.audio_queue.get()

    def stop(self):
        self.stop_event.set()



from __future__ import print_function

from joatmon.assistant.task import BaseTask


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, True, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "download",
            "description": "a function for user to download a file from given url to given path",
            "parameters": {
                "type": "object",
                "properties": {
                    "url": {
                        "type": "string",
                        "description": "url of the file to be downloaded"
                    },
                    "path": {
                        "type": "string",
                        "description": "path of the file to be saved"
                    }
                },
                "required": ["url", "path"]
            }
        }

    def run(self):
        # need to do this in background
        # after it is done, need to notify user and prompt action to continue
        # it should not interfere with the current task that the user running
        if not self.stop_event.is_set():
            self.stop_event.set()



import threading

import cv2


class Camera:
    def __init__(self):
        super(Camera, self).__init__()

        self.cam = None
        self.stop_event = threading.Event()

    def frame(self):
        self.cam = cv2.VideoCapture(0)
        ret, frame = self.cam.read()

        return frame

    def shot(self, path):
        ...

    def stream(self):
        self.cam = cv2.VideoCapture(0)

        while not self.stop_event.is_set():
            ret, frame = self.cam.read()
            yield frame

    def record(self, path, time):
        ...

    def stop(self):
        self.stop_event.set()
        if self.cam is not None:
            self.cam.release()



from contextlib import contextmanager
from threading import Lock

__all__ = ['RWLock']


class RWLock(object):
    def __init__(self):

        self.w_lock = Lock()
        self.num_r_lock = Lock()
        self.num_r = 0

    def r_acquire(self):
        # print('attempting to acquire read lock')
        self.num_r_lock.acquire()
        self.num_r += 1
        if self.num_r == 1:
            self.w_lock.acquire()
        self.num_r_lock.release()

    def r_release(self):
        # print('attempting to release read lock')
        assert self.num_r > 0
        self.num_r_lock.acquire()
        self.num_r -= 1
        if self.num_r == 0:
            self.w_lock.release()
        self.num_r_lock.release()

    @contextmanager
    def r_locked(self):
        try:
            self.r_acquire()
            yield
        finally:
            self.r_release()

    def w_acquire(self):
        # print('attempting to acquire write lock')
        self.w_lock.acquire()

    def w_release(self):
        # print('attempting to release write lock')
        self.w_lock.release()

    @contextmanager
    def w_locked(self):
        try:
            self.w_acquire()
            yield
        finally:
            self.w_release()



from joatmon.plugin.core import Plugin


class DatabasePlugin(Plugin):
    # query can be a dictionary, query builder or formatted query
    # read write strategy
    # when reading, might read on cache so that the read operation will be reduced
    # when writing, might write to message queue and consume in batcher so that the write operation will be reduced
    # after writing the read cache should be updated as well
    async def __aenter__(self):
        await self.start()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            await self.commit()
        else:
            await self.abort()
            raise exc_val
        await self.end()

        return self

    async def create(self, document):
        raise NotImplementedError

    async def alter(self, document):
        raise NotImplementedError

    async def drop(self, document):
        raise NotImplementedError

    async def insert(self, document, *docs):
        raise NotImplementedError

    async def read(self, document, query):
        raise NotImplementedError

    async def update(self, document, query, update):
        raise NotImplementedError

    async def delete(self, document, query):
        raise NotImplementedError

    async def view(self, document, query):
        raise NotImplementedError

    async def execute(self, document, query):
        raise NotImplementedError

    async def count(self, query):
        raise NotImplementedError

    async def start(self):
        raise NotImplementedError

    async def commit(self):
        raise NotImplementedError

    async def abort(self):
        raise NotImplementedError

    async def end(self):
        raise NotImplementedError



import torch.nn as nn


class DQN(nn.Module):
    def __init__(self, in_features, out_features):
        super(DQN, self).__init__()

        self.extractor = nn.Sequential(
            nn.Conv2d(in_channels=in_features, out_channels=32, kernel_size=(8, 8), stride=(4, 4)),
            nn.BatchNorm2d(num_features=32),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels=32, out_channels=64, kernel_size=(4, 4), stride=(2, 2)),
            nn.BatchNorm2d(num_features=64),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=(3, 3), stride=(1, 1)),
            nn.BatchNorm2d(num_features=64),
            nn.ReLU(inplace=True)
        )
        self.predictor = nn.Sequential(
            nn.Linear(in_features=7 * 7 * 64, out_features=512),
            nn.BatchNorm1d(num_features=512),
            nn.ReLU(inplace=True),
            nn.Linear(in_features=512, out_features=out_features),
        )

    def forward(self, x):
        x = self.extractor(x)
        return self.predictor(x.view(x.size(0), -1))



from joatmon.nn import functional as f
from joatmon.nn.core import Optimizer

__all__ = ['Adam']


class Adam(Optimizer):
    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0, amsgrad=False):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, amsgrad=amsgrad)
        super(Adam, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(Adam, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('amsgrad', False)

    def step(self):
        for group in self.param_groups:
            params_with_grad = []
            grads = []
            exp_avgs = []
            exp_avg_sqs = []
            max_exp_avg_sqs = []
            state_steps = []

            for p in group['params']:
                if p.grad is not None:
                    params_with_grad.append(p)
                    grads.append(p.grad)

                    state = self.state[p]
                    if len(state) == 0:
                        state['step'] = 0
                        # Exponential moving average of gradient values
                        state['exp_avg'] = f.zeros_like(p)
                        # Exponential moving average of squared gradient values
                        state['exp_avg_sq'] = f.zeros_like(p)
                        if group['amsgrad']:
                            # Maintains max of all exp. moving avg. of sq. grad. values
                            state['max_exp_avg_sq'] = f.zeros_like(p)

                    exp_avgs.append(state['exp_avg'])
                    exp_avg_sqs.append(state['exp_avg_sq'])

                    if group['amsgrad']:
                        max_exp_avg_sqs.append(state['max_exp_avg_sq'])

                    # update the steps for each param group update
                    state['step'] += 1
                    # record the step after step update
                    state_steps.append(state['step'])

            beta1, beta2 = group['betas']
            f.adam(
                params_with_grad, grads, exp_avgs, exp_avg_sqs, max_exp_avg_sqs,
                state_steps, group['amsgrad'], beta1, beta2, group['lr'], group['weight_decay'], group['eps']
            )



from joatmon.plugin.core import Plugin


class MessagePlugin(Plugin):
    def get_producer(self, topic):
        raise NotImplementedError

    def get_consumer(self, topic):
        raise NotImplementedError



from typing import Union

from joatmon import context
from joatmon.orm.enum import Enum
from joatmon.plugin.core import Plugin


class LogLevel(Enum):
    NotSet = 1 << 0
    Debug = 1 << 1
    Info = 1 << 2
    Warning = 1 << 3
    Error = 1 << 4
    Critical = 1 << 5


class LoggerPlugin(Plugin):
    def __init__(self, level, language, ip):
        if isinstance(level, LogLevel):
            level = level.name
        self._level = LoggerPlugin._get_level(level)

        self.language = language
        self.ip = ip

    @staticmethod
    def _get_level(level_str):
        _level_key = list(filter(lambda x: level_str.lower() == x.lower(), LogLevel.__dict__.keys()))
        if len(_level_key) == 1:
            _level_key = _level_key[0]
        else:
            _level_key = LogLevel.Info.name

        return LogLevel[_level_key]

    async def _write(self, log: dict):
        raise NotImplementedError

    async def log(self, log: dict, level: Union[LogLevel, str] = LogLevel.Debug):
        try:
            if isinstance(level, str):
                level = LoggerPlugin._get_level(level)

            if self._level.value <= level.value:
                log['level'] = level.name
                log['language'] = context.get_value(self.language).get()
                log['ip'] = context.get_value(self.ip).get()
                await self._write(log)
        except Exception as ex:
            print(str(ex))

    async def debug(self, log):
        await self.log(log=log, level=LogLevel.Debug)

    async def info(self, log):
        await self.log(log=log, level=LogLevel.Info)

    async def warning(self, log):
        await self.log(log=log, level=LogLevel.Warning)

    async def error(self, log):
        await self.log(log=log, level=LogLevel.Error)

    async def critical(self, log):
        await self.log(log=log, level=LogLevel.Critical)



import sys
from enum import (
    EnumMeta,
    IntEnum
)

from joatmon.utility import (
    to_pascal_string,
    to_snake_string
)


class Meta(EnumMeta):
    def __contains__(self, item):
        try:
            self(item)
        except ValueError:
            return False
        else:
            return True


class Enum(IntEnum, metaclass=Meta):
    def __int__(self):
        return self.value

    def __str__(self):
        return f'{to_snake_string(type(self).__name__.replace("Enum", ""))}.{self.name}'
        # return '{@resource.' + f'{to_snake_string(type(self).__name__.replace("Enum", ""))}.{self.name}' + '}'

    def __repr__(self):
        return f'{type(self).__name__}.{self.name}'

    def __lt__(self, other):
        if other not in self.__class__:
            raise ValueError(f'given value: {other} could not be casted to {self.__class__}')

        if not isinstance(other, int):
            other = int(other)
        return int(self) < other

    def __gt__(self, other):
        if other not in self.__class__:
            raise ValueError(f'given value: {other} could not be casted to {self.__class__}')

        if not isinstance(other, int):
            other = int(other)
        return int(self) > other

    def __eq__(self, other):
        if other not in self.__class__:
            raise ValueError(f'given value: {other} could not be casted to {self.__class__}')

        if not isinstance(other, int):
            other = int(other)
        return int(self) == other

    @staticmethod
    def parse(value: str):
        type_name, attribute_name = value.split('.')
        type_name = to_pascal_string(type_name)

        _type = getattr(sys.modules[__name__], f'{type_name}Enum', None)
        if _type is not None:
            return _type[attribute_name]



from elevenlabs import play


class Speaker:
    def __init__(self):
        super(Speaker, self).__init__()

        # set_api_key(json.loads(open('iva/iva.json', 'r').read())['config']['elevenlabs']['key'])

    def say(self, audio):
        play(audio)



import pytest


def test_j_s_o_n_decoder_object_hook():
    assert True is True


def test_j_s_o_n_encoder_default():
    assert True is True


def test_current_time():
    from joatmon.utility import current_time

    current_time()

    assert True is True


def test_empty_object_id():
    import uuid
    from joatmon.utility import empty_object_id

    assert empty_object_id() != uuid.uuid4()


def test_first():
    from joatmon.utility import first

    assert first([1, 2, 3]) == 1


def test_first_async():
    import asyncio
    from joatmon.utility import first_async

    async def t():
        async def t_():
            yield 1
            yield 2

        await first_async(t_())

    asyncio.ensure_future(t())

    assert True is True


def test_get_class_that_defined_method():
    from joatmon.utility import (
        JSONEncoder,
        get_class_that_defined_method
    )

    get_class_that_defined_method(JSONEncoder.default)

    assert True is True


def test_get_converter():
    import uuid
    import datetime
    from joatmon.utility import get_converter
    from joatmon.serializable import Serializable

    assert get_converter(int)(1.0) == 1
    assert get_converter(int)(1) == 1
    assert get_converter(int)('1') == 1
    assert get_converter(datetime.datetime)(datetime.datetime(2022, 1, 1).isoformat()).year == 2022
    assert get_converter(datetime.datetime)(datetime.datetime(2022, 1, 1)).year == 2022
    assert get_converter(float)(1.0) == 1
    assert get_converter(float)('1.0') == 1
    assert get_converter(float)(1) == 1
    assert get_converter(str)('1') == '1'
    assert get_converter(bytes)(b'1') == b'1'
    assert get_converter(bool)(True) is True
    assert get_converter(uuid.UUID)(uuid.uuid4())
    assert get_converter(uuid.UUID)(str(uuid.uuid4()))
    assert get_converter(dict)({}) == {}
    assert get_converter(dict)(Serializable()) == {}
    assert get_converter(list)([]) == []
    assert get_converter(tuple)(()) == ()
    assert get_converter(object)({}) == {}
    assert get_converter(object)([]) == []
    assert get_converter(object)(()) == ()
    assert get_converter(object)('') == ''
    assert get_converter(object)(Serializable()) == {}


def test_get_function_args():
    from joatmon.utility import (
        get_function_args,
        JSONEncoder
    )

    def t():
        ...

    get_function_args(JSONEncoder, '12')
    get_function_args(t)

    assert True is True


def test_get_function_kwargs():
    from joatmon.utility import get_function_kwargs

    def t():
        ...

    get_function_kwargs(t)

    assert True is True


def test_get_module_functions():
    import joatmon
    from joatmon.utility import get_module_functions

    get_module_functions(joatmon)

    assert True is True


def test_ip_validator():
    from joatmon.utility import ip_validator

    assert ip_validator('127.0.0.1')


def test_mail_validator():
    from joatmon.utility import mail_validator

    assert mail_validator('hamitcanmalkoc@gmail.com')


def test_new_nickname():
    from joatmon.utility import new_nickname

    assert new_nickname() != ''


def test_new_object_id():
    from joatmon.utility import (
        new_object_id,
        empty_object_id
    )

    assert new_object_id() != empty_object_id()


def test_new_password():
    from joatmon.utility import new_password

    assert new_password() != ''


def test_single():
    from joatmon.utility import single

    assert single([1]) == 1


def test_single_async():
    import asyncio
    from joatmon.utility import single_async

    async def t():
        async def t_():
            yield 1

        await single_async(t_())

    asyncio.ensure_future(t())

    assert True is True


def test_to_case():
    from joatmon.utility import to_case

    assert to_case('snake', {'A': 1}) == {'a': 1}
    assert to_case('pascal', {'a': 1}) == {'A': 1}
    assert to_case('upper', {'a': 1}) == {'A': 1}
    assert to_case('lower', {'A': 1}) == {'a': 1}


def test_to_enumerable():
    from joatmon.serializable import Serializable
    from joatmon.utility import to_enumerable

    assert to_enumerable([1, 2]) == [1, 2]
    assert to_enumerable((1, 2)) == (1, 2)
    assert to_enumerable(Serializable(a=1, b=2)) == {'a': 1, 'b': 2}


def test_to_hash():
    from joatmon.utility import to_hash

    def t(a, b):
        ...

    to_hash(t, ('1',), {'b': 2})

    assert True is True


def test_to_list():
    from joatmon.utility import to_list

    assert to_list([1, 2]) == [1, 2]


def test_to_list_async():
    import asyncio
    from joatmon.utility import to_list_async

    async def t():
        async def t_():
            yield 1
            yield 2

        await to_list_async(t_())

    asyncio.ensure_future(t())

    assert True is True


def test_to_lower_string():
    from joatmon.utility import to_lower_string

    assert to_lower_string('Abc') == 'abc'


def test_to_pascal_string():
    from joatmon.utility import to_pascal_string

    assert to_pascal_string('abc') == 'Abc'


def test_to_snake_string():
    from joatmon.utility import to_snake_string

    assert to_snake_string('ABC') == 'a_b_c'


def test_to_title():
    from joatmon.utility import to_title

    assert to_title('a') == 'A'


def test_to_upper_string():
    from joatmon.utility import to_upper_string

    assert to_upper_string('a') == 'A'


if __name__ == '__main__':
    pytest.main([__file__])



import math
from typing import (
    List,
    Optional,
    Tuple,
    Union
)

from joatmon.array.array import Array

__all__ = [
    'absolute', 'add', 'amax', 'maximum', 'amin', 'minimum', 'arange', 'argmax', 'argmin', 'around', 'array', 'asarray', 'astype', 'ceil',
    'clip', 'concatenate', 'copy', 'dim', 'dot', 'empty', 'eq', 'exp', 'expand_dims', 'eye', 'fill', 'flatten',
    'floor', 'floordiv', 'full', 'ge', 'getitem', 'gt', 'indices', 'le', 'linspace', 'lt', 'mean',
    'median', 'mul', 'ne', 'negative', 'ones', 'ones_like', 'pad', 'power', 'prod', 'put_along_axis', 'repeat',
    'reshape', 'setitem', 'size', 'split', 'sqrt', 'square', 'squeeze', 'stack', 'std', 'sub', 'sum',
    'take_along_axis', 'tanh', 'tolist', 'transpose', 'truediv', 'unique', 'var', 'where', 'zeros', 'zeros_like'
]

BooleanT = Union[bool]
IntegerT = Union[int]
FloatingT = Union[float]
NumberT = Union[IntegerT, FloatingT]

DataT = Union[BooleanT, IntegerT, FloatingT]
ArrayT = Union[Tuple[DataT], List[DataT], Array, Tuple['ArrayT'], List['ArrayT']]
TypeT = Union[str]
ShapeT = Union[Tuple[int], List[int]]
IndexT = Union[int, slice, Tuple[Union[int, slice], ...]]
AxisT = Optional[Union[int, Tuple[int], List[int]]]


class Broadcast:
    def __init__(self, inp1: Union[DataT, ArrayT], inp2: Union[DataT, ArrayT]):
        # both parameters are float, int, bool
        if dim(inp1) == 0 and dim(inp2) == 0:
            pass

        # first parameter is NDArray, second parameter is float, int, bool
        if dim(inp1) != 0 and dim(inp2) == 0:
            inp2 = reshape(repeat(inp2, prod(size(inp1))), size(inp1))

        # both parameters are NDArray
        if dim(inp1) != 0 and dim(inp2) != 0 and dim(inp1) != dim(inp2):
            # print('reshaping inp2 array')
            new_shape = [1] * (len(size(inp1)) - len(size(inp2))) + list(size(inp2))
            inp2 = reshape(inp2, new_shape)

        # first dimension sizes are the same, it can be broadcast together
        if dim(inp1) != 0 and dim(inp2) != 0 and size(inp1, 0) == size(inp2, 0):
            pass

        # first dimension sizes are not the same and first dimension size of the second input is not 1
        # print(inp1.shape, inp2.shape)
        if dim(inp1) != 0 and dim(inp2) != 0 and size(inp1, 0) != size(inp2, 0) and size(inp2, 0) != 1:
            raise ValueError('cannot broadcast')

        # first dimension sizes are not the same but the first dimension size of the second input is 1, it can be broadcast
        if dim(inp1) != 0 and dim(inp2) != 0 and size(inp1, 0) != size(inp2, 0) and size(inp2, 0) == 1:
            # print('repeating and reshaping inp2 array')
            inp2 = reshape(repeat(inp2, size(inp1, 0)), (size(inp1, 0),) + size(inp2)[1:])

        self.inp1 = inp1
        self.inp2 = inp2

        if dim(self.inp1) != 0 and dim(self.inp2) != 0:
            self.zip = zip(self.inp1, self.inp2)
        else:
            raise ValueError('an error has happened')

    def __iter__(self):
        return self.zip

    def __next__(self):
        return next(self.zip)

    def reset(self):
        if dim(self.inp1) != 0 and dim(self.inp2) != 0:
            self.zip = zip(self.inp1, self.inp2)
        else:
            raise ValueError('an error has happened')


def array(arr: ArrayT) -> ArrayT:
    return Array(arr)


def asarray(arr: ArrayT) -> ArrayT:
    return Array(arr)


def arange(start: float, stop: float, step: float) -> Array:
    return linspace(start, stop, int((stop - start) // step))


def linspace(start: float, stop: float, steps: int) -> Array:
    inc = (stop - start) / steps
    return Array([start + inc * idx for idx in range(steps)])


def eye(rows: int, columns: int) -> Array:
    # need to check which one is smaller
    ret = zeros((rows, columns))
    for row_col in range(min(rows, columns)):
        ret[row_col][row_col] = 1
    return ret


def empty(shape) -> Array:
    if len(shape) == 0:
        raise ValueError('array shape has to be at least 1-dimensional')

    ret = []
    for _ in range(shape[0]):
        ret.append(tolist(zeros(shape[1:])))
    return Array(ret)


def full(shape) -> Array:
    if len(shape) == 0:
        raise ValueError('array shape has to be at least 1-dimensional')

    ret = []
    for _ in range(shape[0]):
        ret.append(tolist(ones(shape[1:])))
    return Array(ret)


def zeros(shape) -> Array:
    if len(shape) == 0:
        raise ValueError('array shape has to be at least 1-dimensional')

    ret = []
    for _ in range(shape[0]):
        if len(shape[1:]) == 0:
            ret.append(float(0))
        else:
            ret.append(tolist(zeros(shape[1:])))
    return Array(ret)


def ones(shape) -> Array:
    if len(shape) == 0:
        raise ValueError('array shape has to be at least 1-dimensional')

    ret = []
    for _ in range(shape[0]):
        ret.append(tolist(ones(shape[1:])))
    return Array(ret)


def ones_like(inp: ArrayT) -> Array:
    return ones(size(inp))


def zeros_like(inp: ArrayT) -> Array:
    return zeros(size(inp))


def concatenate(inputs, axis: AxisT = None) -> ArrayT:
    ...


def stack(inputs, axis: AxisT = None) -> ArrayT:
    ...


def astype(inp: Union[DataT, ArrayT], dtype: TypeT) -> Union[DataT, ArrayT]:
    if isinstance(inp, (bool, int, float)):
        if dtype in ('bool', 'boolean'):
            return bool(inp)
        if dtype in ('int', 'integer', 'int8', 'int16', 'int32', 'int64'):
            return int(inp)
        if dtype in ('float', 'floating', 'float8', 'float16', 'float32', 'float64'):
            return float(inp)
    elif isinstance(inp, (tuple, list, Array)):
        ret = []
        for data in inp:
            ret.append(astype(data, dtype))
        return type(inp)(ret)
    else:
        raise ValueError('value type is not recognized')


def copy(inp: Union[DataT, ArrayT]) -> Union[DataT, ArrayT]:
    if isinstance(inp, (bool, int, float)):
        return type(inp)(inp)
    elif isinstance(inp, (tuple, list, Array)):
        return type(inp)(tolist(inp))
    else:
        raise ValueError(f'cannot copy object of type {type(inp)}')


def repeat(inp: Union[DataT, ArrayT], count: DataT, axis=0) -> ArrayT:
    # need to use axis
    # need to be able to return tuple or list

    ret = []
    for _ in range(count):
        if isinstance(inp, (bool, int, float)):
            ret.append(inp)
        elif isinstance(inp, (tuple, list, Array)):
            ret.append(tolist(inp))
    return Array(ret)


def split(inp: ArrayT, chunks, axis=0) -> ArrayT:
    # need to use axis
    # need to return tuple

    length = len(inp)
    chunk_length = length // chunks
    if chunk_length * chunks != length:
        raise ValueError(f'array with length {length} cannot be divided into {chunks} chunks')

    ret = []
    for chunk in range(chunks):
        ret.append(inp[chunk * chunk_length: (chunk + 1) * chunk_length])
    return type(inp)(ret)


def tolist(inp: ArrayT) -> list:
    ret = []
    for data in inp:
        if isinstance(data, (bool, int, float)):
            ret.append(data)
        elif isinstance(data, (tuple, list, Array)):
            ret.append(tolist(data))
        else:
            raise ValueError(f'{type(data)} could not be converted')
    return ret


def getitem(inp: ArrayT, idx: IndexT):
    if isinstance(inp, (tuple, list)):
        if isinstance(idx, int):
            return copy(inp[idx])
        elif isinstance(idx, slice):
            ret = []
            for data in inp[idx]:
                ret.append(data)
            return type(inp)(ret)
        elif isinstance(idx, (tuple, list)) and len(idx) == 1:
            ret = inp[idx[0]]
            if isinstance(ret, (bool, int, float)):
                return ret
            elif isinstance(ret, (tuple, list, Array)):
                return type(inp)(ret)
        elif isinstance(idx, (tuple, list)):
            ret = []
            if isinstance(idx[0], int):
                return getitem(inp[idx[0]], idx[1:])
            elif isinstance(idx[0], slice):
                for data in inp[idx[0]]:
                    ret.append(getitem(data, idx[1:]))
            return type(inp)(ret)
    elif isinstance(inp, Array):
        ret = getitem(tolist(inp), idx)
        if isinstance(ret, (bool, int, float)):
            return ret
        return Array(ret)
    else:
        raise ValueError(f'object type {type(inp)} is not recognized')


def take_along_axis(inp: Union[DataT, ArrayT], indexes, axis) -> ArrayT:
    ...


def setitem(inp: ArrayT, idx: IndexT, value: Union[DataT, ArrayT]):
    # need to repeat inp[idx].shape times
    if isinstance(value, (bool, int, float)):
        inp_shape = size(inp)
        inp_size = prod(inp_shape)
        new_value = reshape(repeat(value, inp_size), inp_shape)
    elif isinstance(value, (tuple, list)):
        if size(inp) != size(value):
            raise ValueError(f'{size(inp)} and {size(value)} does not match')
        new_value = copy(value)
    elif isinstance(value, Array):
        if size(inp) != size(value):
            raise ValueError(f'{size(inp)} and {size(value)} does not match')
        new_value = copy(value)
    else:
        raise ValueError(f'{type(value)} is not recognized as value type')

    if isinstance(inp, (tuple, list)):
        if isinstance(idx, int):
            inp[idx] = new_value[idx]
        elif isinstance(idx, slice):
            for _idx in range(idx.start, idx.stop, idx.step):
                inp[_idx] = new_value[_idx]
        elif isinstance(idx, (tuple, list)) and len(idx) == 1:
            inp[idx[0]] = new_value[idx[0]]
        elif isinstance(idx, (tuple, list)):
            if isinstance(idx[0], int):
                setitem(inp[idx[0]], idx[1:], new_value[idx[0]])
            elif isinstance(idx[0], slice):
                for _idx in range(idx[0].start, idx[0].stop, idx[0].step):
                    setitem(inp[_idx], idx[1:], new_value[_idx])
            else:
                raise ValueError(f'{type(idx[0])} type is not recognized as index')
        else:
            raise ValueError(f'{type(idx)} type is not recognized as index')
        # does not work when not returning
    elif isinstance(inp, Array):
        list_inp = tolist(inp)
        setitem(list_inp, idx, new_value)
        return Array(list_inp)
    else:
        raise ValueError(f'object type {type(inp)} is not recognized')


def put_along_axis(inp: Union[DataT, ArrayT], indexes, values, axis) -> ArrayT:
    ...


def where(condition) -> ArrayT:
    ...


def indices(dimensions) -> ArrayT:
    ...


def dim(inp: Union[DataT, ArrayT]) -> int:
    if isinstance(inp, (bool, int, float)):
        return 0
    return dim(inp[0]) + 1


def size(inp: Union[DataT, ArrayT], axis: AxisT = None):
    if isinstance(inp, (bool, int, float)):
        return ()

    if axis is None or axis < 0:
        return tuple([len(inp)] + list(size(inp[0])))
    return tuple((size(inp[0], axis=axis - 1)))


def flatten(inp: ArrayT) -> ArrayT:
    data_type = type(inp)

    ret = []
    for data in inp:
        if dim(data) == 0:
            ret.append(data)
        elif dim(data) == 1:
            ret += tolist(data)
        else:
            ret += tolist(flatten(data))
    return data_type(ret)


def reshape(inp: ArrayT, shape) -> ArrayT:
    flat = flatten(inp)

    subdims = shape[1:]
    subsize = prod(Array(subdims))
    if shape[0] * subsize != len(flat):
        raise ValueError('size does not match or invalid')
    if not subdims:
        return flat
    return Array([reshape(flat[i: i + subsize], subdims) for i in range(0, len(flat), subsize)])


def squeeze(inp: ArrayT, axis: AxisT = None) -> ArrayT:
    if axis is None:
        if any([x == 1 for x in size(inp)]):
            new_shape = tuple(filter(lambda x: x != 1, size(inp)))
        else:
            new_shape = size(inp)
    else:
        axis_size = size(inp, axis)
        if axis_size != 1:
            raise ValueError('cannot select an axis to squeeze out which has size not equal to one')
        new_shape = list(size(inp))
        new_shape.pop(axis)

    return reshape(inp, new_shape)


def expand_dims(inp: ArrayT, axis) -> ArrayT:
    shape = list(size(inp))
    shape.insert(axis, 1)

    return reshape(inp, shape)


def pad(inp: Union[DataT, ArrayT], padding, mode) -> ArrayT:
    ...


def transpose(inp: Union[DataT, ArrayT], axes=None) -> ArrayT:
    def multi_to_one(idx, s):
        r = 0
        for i, _idx in enumerate(idx):
            r += (prod(s[i + 1:]) * _idx)
        return r

    def one_to_multi(idx, s):
        r = []
        for i in range(len(s)):
            div, mod = divmod(idx, prod(s[i + 1:]))
            r.append(div)
            idx = mod
        return r

    if axes is None:
        if dim(inp) != 2:
            raise ValueError(f'axes has to be used on {dim(inp)}-dim array')
        axes = (1, 0)

    shape = size(inp)
    new_shape = []
    for axis in axes:
        new_shape.append(shape[axis])
    new_shape = tuple(new_shape)

    new_flatten = flatten(zeros(new_shape))
    old_flatten = flatten(inp)

    for flat_idx in range(len(old_flatten)):
        old_multi = one_to_multi(flat_idx, shape)
        new_multi = []
        for axis in axes:
            new_multi.append(old_multi[axis])

        old_flat_idx = multi_to_one(old_multi, shape)
        new_flat_idx = multi_to_one(new_multi, new_shape)

        new_flatten = setitem(new_flatten, new_flat_idx, getitem(old_flatten, old_flat_idx))

    return reshape(new_flatten, new_shape)


def fill(inp: ArrayT, value: DataT) -> ArrayT:
    shape = size(inp)
    flat_array = flatten(inp)
    for idx in range(len(flat_array)):
        setitem(flat_array, idx, value)
    return reshape(flat_array, shape)


def absolute(inp: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    flat_array = flatten(inp)
    for idx in range(len(flat_array)):
        setitem(flat_array, idx, abs(getitem(flat_array, idx)))
    return reshape(flat_array, size(inp))


def negative(inp: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    flat_array = flatten(inp)
    for idx in range(len(flat_array)):
        setitem(flat_array, idx, -getitem(flat_array, idx))
    return reshape(flat_array, size(inp))


def around(inp: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    flat_array = flatten(inp)
    for idx in range(len(flat_array)):
        setitem(flat_array, idx, round(getitem(flat_array, idx)))
    return reshape(flat_array, size(inp))


def floor(inp: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    flat_array = flatten(inp)
    for idx in range(len(flat_array)):
        setitem(flat_array, idx, math.floor(getitem(flat_array, idx)))
    return reshape(flat_array, size(inp))


def ceil(inp: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    flat_array = flatten(inp)
    for idx in range(len(flat_array)):
        setitem(flat_array, idx, math.ceil(getitem(flat_array, idx)))
    return reshape(flat_array, size(inp))


def sqrt(inp: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    flat_array = flatten(inp)
    for idx in range(len(flat_array)):
        setitem(flat_array, idx, math.sqrt(getitem(flat_array, idx)))
    return reshape(flat_array, size(inp))


def square(inp: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    flat_array = flatten(inp)
    for idx in range(len(flat_array)):
        setitem(flat_array, idx, getitem(flat_array, idx) ** 2)
    return reshape(flat_array, size(inp))


def clip(inp: Union[DataT, ArrayT], min_value: DataT, max_value: DataT, *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    flat_array = flatten(inp)
    for idx in range(len(flat_array)):
        value = getitem(flat_array, idx)
        if value < min_value:
            value = min_value
        if value > max_value:
            value = max_value
        setitem(flat_array, idx, value)
    return reshape(flat_array, size(inp))


def exp(inp: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    flat_array = flatten(inp)
    for idx in range(len(flat_array)):
        setitem(flat_array, idx, math.e ** getitem(flat_array, idx))
    return reshape(flat_array, size(inp))


def tanh(inp: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    flat_array = flatten(inp)
    for idx in range(len(flat_array)):
        setitem(flat_array, idx, math.tanh(getitem(flat_array, idx)))
    return reshape(flat_array, size(inp))


def sum(inp: ArrayT, axis: AxisT = None, keepdims=False) -> Union[DataT, ArrayT]:
    if axis is None:
        if dim(inp) == 0:
            return astype(copy(inp), 'float')

        ret = float(0)
        for idx in range(len(inp)):
            ret += sum(inp[idx], axis=axis)
        return ret

    # kahan sum
    shape = size(inp)
    s = zeros(shape[:axis] + shape[axis + 1:])
    c = zeros(size(s))
    for i in range(shape[axis]):
        # y = getitem(inp, (slice(None),) * axis + (i,)) - c  # inp[(slice(None),) * axis + (i,)] - c
        y = inp[(slice(None),) * axis + (i,)] - c
        t = s + y
        c = (t - s) - y
        s = copy(t)
    return s


def mean(inp: ArrayT, axis: AxisT = None) -> Union[DataT, ArrayT]:
    shape = size(inp)
    s = sum(inp, axis)
    n = shape[axis] if axis is not None else prod(shape)
    return s / n


def median(inp: ArrayT, axis: AxisT = None) -> Union[DataT, ArrayT]:
    ...


def var(inp: ArrayT, axis: AxisT = None) -> Union[DataT, ArrayT]:
    if axis is not None:
        m_shape = list(size(inp))
        m_shape[axis] = 1
        m = reshape(mean(inp, axis), m_shape)
        a = absolute(inp - m)
        return mean(a ** 2, axis=axis)

    return mean(absolute(inp - mean(inp)) ** 2)


def std(inp: ArrayT, axis: AxisT = None) -> Union[DataT, ArrayT]:
    return sqrt(var(inp, axis=axis))


def prod(inp: ArrayT, axis: AxisT = None) -> int:
    # could be multi dimensional
    p = 1
    if isinstance(inp, (tuple, list)):
        for data in inp:
            p *= data
    elif isinstance(inp, Array):
        return prod(inp.data)
    return p


def unique(inp: ArrayT, return_counts=False) -> ArrayT:
    ...


def argmax(inp: ArrayT, axis: AxisT = None) -> Union[DataT, ArrayT]:
    ...


def argmin(inp: ArrayT, axis: AxisT = None) -> Union[DataT, ArrayT]:
    ...


def amax(inp: ArrayT, axis: AxisT = None) -> Union[DataT, ArrayT]:
    ...


def maximum(inp: ArrayT, axis: AxisT = None) -> Union[DataT, ArrayT]:
    ...


def amin(inp: ArrayT, axis: AxisT = None) -> Union[DataT, ArrayT]:
    ...


def minimum(inp: ArrayT, axis: AxisT = None) -> Union[DataT, ArrayT]:
    ...


def add(inp1: Union[DataT, ArrayT], inp2: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    if dim(inp1) == 0 and dim(inp2) == 0:
        return inp1 + inp2

    ret = out or zeros(size(inp1))
    broadcast = Broadcast(inp1, inp2)

    for idx, (_inp1, _inp2) in enumerate(broadcast):
        ret[idx] = add(_inp1, _inp2, out=ret[idx])
    return ret


def sub(inp1: Union[DataT, ArrayT], inp2: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    if dim(inp1) == 0 and dim(inp2) == 0:
        return inp1 - inp2

    ret = out or zeros(size(inp1))
    broadcast = Broadcast(inp1, inp2)

    for idx, (_inp1, _inp2) in enumerate(broadcast):
        ret[idx] = sub(_inp1, _inp2, out=ret[idx])
    return ret


def mul(inp1: Union[DataT, ArrayT], inp2: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    if dim(inp1) == 0 and dim(inp2) == 0:
        return inp1 * inp2

    ret = out or zeros(size(inp1))
    broadcast = Broadcast(inp1, inp2)

    for idx, (_inp1, _inp2) in enumerate(broadcast):
        ret[idx] = mul(_inp1, _inp2, out=ret[idx])
    return ret


def truediv(inp1: Union[DataT, ArrayT], inp2: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    if dim(inp1) == 0 and dim(inp2) == 0:
        return inp1 / inp2

    ret = out or zeros(size(inp1))
    broadcast = Broadcast(inp1, inp2)

    for idx, (_inp1, _inp2) in enumerate(broadcast):
        ret[idx] = truediv(_inp1, _inp2, out=ret[idx])
    return ret


def floordiv(inp1: Union[DataT, ArrayT], inp2: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    if dim(inp1) == 0 and dim(inp2) == 0:
        return inp1 // inp2

    ret = out or zeros(size(inp1))
    broadcast = Broadcast(inp1, inp2)

    for idx, (_inp1, _inp2) in enumerate(broadcast):
        ret[idx] = floordiv(_inp1, _inp2, out=ret[idx])
    return ret


def power(inp: Union[DataT, ArrayT], p: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    if dim(inp) == 0:
        return inp ** p

    ret = out or zeros(size(inp))
    for idx, _inp in enumerate(inp):
        ret[idx] = power(_inp, p, out=ret[idx])
    return ret


def lt(inp1: Union[DataT, ArrayT], inp2: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    if dim(inp1) == 0 and dim(inp2) == 0:
        return inp1 < inp2

    ret = out or zeros(size(inp1))
    broadcast = Broadcast(inp1, inp2)

    for idx, (_inp1, _inp2) in enumerate(broadcast):
        ret[idx] = lt(_inp1, _inp2, out=ret[idx])
    return ret


def le(inp1: Union[DataT, ArrayT], inp2: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    if dim(inp1) == 0 and dim(inp2) == 0:
        return inp1 <= inp2

    ret = out or zeros(size(inp1))
    broadcast = Broadcast(inp1, inp2)

    for idx, (_inp1, _inp2) in enumerate(broadcast):
        ret[idx] = le(_inp1, _inp2, out=ret[idx])
    return ret


def gt(inp1: Union[DataT, ArrayT], inp2: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    if dim(inp1) == 0 and dim(inp2) == 0:
        return inp1 > inp2

    ret = out or zeros(size(inp1))
    broadcast = Broadcast(inp1, inp2)

    for idx, (_inp1, _inp2) in enumerate(broadcast):
        ret[idx] = gt(_inp1, _inp2, out=ret[idx])
    return ret


def ge(inp1: Union[DataT, ArrayT], inp2: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    if dim(inp1) == 0 and dim(inp2) == 0:
        return inp1 >= inp2

    ret = out or zeros(size(inp1))
    broadcast = Broadcast(inp1, inp2)

    for idx, (_inp1, _inp2) in enumerate(broadcast):
        ret[idx] = ge(_inp1, _inp2, out=ret[idx])
    return ret


def eq(inp1: Union[DataT, ArrayT], inp2: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    if dim(inp1) == 0 and dim(inp2) == 0:
        return inp1 == inp2

    ret = out or zeros(size(inp1))
    broadcast = Broadcast(inp1, inp2)

    for idx, (_inp1, _inp2) in enumerate(broadcast):
        ret[idx] = eq(_inp1, _inp2, out=ret[idx])
    return ret


def ne(inp1: Union[DataT, ArrayT], inp2: Union[DataT, ArrayT], *, out: Optional[ArrayT] = None) -> Union[DataT, ArrayT]:
    if dim(inp1) == 0 and dim(inp2) == 0:
        return inp1 != inp2

    ret = out or zeros(size(inp1))
    broadcast = Broadcast(inp1, inp2)

    for idx, (_inp1, _inp2) in enumerate(broadcast):
        ret[idx] = ne(_inp1, _inp2, out=ret[idx])
    return ret


def dot(inp1: ArrayT, inp2: ArrayT, *, out: Optional[ArrayT] = None) -> ArrayT:
    ...



import functools
import inspect

from joatmon import context


def authorized(auth, token, issuer):  # use current token and issuer
    def _decorator(func):
        @functools.wraps(func)
        async def _wrapper(*args, **kwargs):
            token_value = context.get_value(token).get()
            issuer_value = context.get_value(issuer).get()

            audience = f'{func.__module__}.{func.__qualname__}'

            authorizer = context.get_value(auth)
            await authorizer.authorize(token_value, issuer_value, audience)

            return await func(*args, **kwargs)

        _wrapper.__signature__ = inspect.signature(func)  # need to set min role here or maybe new decorator
        return _wrapper

    return _decorator



from __future__ import print_function

import os
import platform
import sys
from distutils.core import setup

from setuptools import find_packages

from joatmon.version import get_version

extras = {
    'ai': ['numpy', 'torch', 'matplotlib', 'six', 'Pillow'],
    'algorithm': [],
    'assistant': ['feedparser', 'requests', 'colorama', 'schedule', 'psutil', 'nltk', 'tensorflow', 'openai', 'transitions'],
    'content': [],
    'core': [],
    'decorator': ['async_exit_stack'],
    'download': ['pycurl'],
    'event': [],
    'game': ['gym', 'pygame', 'opencv-contrib-python', 'pymunk'],
    'hid': ['whisper', 'pyttsx3'],
    'orm': [],
    'plugin': ['pyjwt', 'redis', 'couchbase', 'elasticsearch', 'pymongo', 'psycopg2', 'pykafka', 'pyotp'],
    'search': ['requests'],
    'structure': [],
    'system': ['opencv-contrib-python', 'numpy', 'elevenlabs', 'openai'],
}
if sys.platform == 'win32':
    extras['system'].extend(['pywin32'])
extras['all'] = list(set([item for group in extras.values() for item in group]))

if sys.version_info < (3,):
    print("Python 2 has reached end-of-life and is no longer supported by joatmon.")
    sys.exit(-1)
if sys.platform == 'win32' and sys.maxsize.bit_length() == 31:
    print("32-bit Windows Python runtime is not supported. Please switch to 64-bit Python.")
    sys.exit(-1)

python_min_version = (3, 9, 1)
python_min_version_str = '.'.join(map(str, python_min_version))
if sys.version_info < python_min_version:
    print("You are using Python {}. Python >={} is required.".format(platform.python_version(), python_min_version_str))
    sys.exit(-1)

if __name__ == '__main__':
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "README.md"), encoding="utf-8") as f:
        long_description = f.read()

    version_range_max = max(sys.version_info[1], 8) + 1
    setup(
        name=os.getenv('JOATMON_PACKAGE_NAME', 'joatmon'),
        version=get_version(),
        description="General Purpose Python Library",
        long_description=long_description,
        long_description_content_type="text/markdown",
        packages=[package for package in find_packages(exclude='tests')],
        entry_points={
            'console_scripts': [
                'assistant = scripts.assistant:main'
            ]
        },
        install_requires=[],
        tests_require=[
            'pytest',
        ],
        extras_require=extras,
        package_data={
            'joatmon': [
                'game/assets/chess/*.png',
                'game/assets/sokoban/sprites/*.png',
                'game/assets/sokoban/xmls/*.xml'
            ]
        },
        url='https://github.com/malkoch/joatmon',
        download_url='https://github.com/malkoch/joatmon/tags',
        author='Hamitcan Malkoç',
        author_email='hamitcanmalkoc@gmail.com',
        python_requires='>={}'.format(python_min_version_str),
        classifiers=[
                        'Development Status :: 5 - Production/Stable',
                        'Intended Audience :: Developers',
                        'Topic :: Software Development',
                        'Topic :: Software Development :: Libraries',
                        'Topic :: Software Development :: Libraries :: Python Modules',
                        'Programming Language :: Python :: 3',
                    ] + ['Programming Language :: Python :: 3.{}'.format(i) for i in range(python_min_version[1], version_range_max)],
        license='MIT',
        keywords='joatmon ml idm automation iva',
    )



import asyncio
from struct import (
    pack,
    unpack
)

from .errors import NoDataException


def bytes_to_hex_str(data):
    return '0x' + ''.join(format(b, '02x') for b in data)


def bytes_to_int(data):
    try:
        return int.from_bytes(data, byteorder='big')
    except:
        return data


def int_to_bytes(int_value: int, length: int) -> bytes:
    if length == 1:
        fmt = "!B"
    elif length == 2:
        fmt = "!H"
    else:
        raise ValueError('fmt has to be specified, length has to be either 1 or 2')
    return pack(fmt, int_value)


async def read_or_raise(reader, n=-1):
    try:
        data = await reader.read(n)
    except (asyncio.IncompleteReadError, ConnectionResetError, BrokenPipeError) as ex:
        data = None
    if not data:
        raise NoDataException("No more data")
    return data


async def decode_string(reader) -> str:
    length_bytes = await read_or_raise(reader, 2)
    str_length = unpack("!H", length_bytes)
    if str_length[0]:
        byte_str = await read_or_raise(reader, str_length[0])
        try:
            return byte_str.decode(encoding='utf-8')
        except:
            return str(byte_str)
    else:
        return ''


async def decode_data_with_length(reader) -> bytes:
    length_bytes = await read_or_raise(reader, 2)
    bytes_length = unpack("!H", length_bytes)
    data = await read_or_raise(reader, bytes_length[0])
    return data


def encode_string(string: str) -> bytes:
    data = string.encode(encoding='utf-8')
    data_length = len(data)
    return int_to_bytes(data_length, 2) + data


def encode_data_with_length(data: bytes) -> bytes:
    data_length = len(data)
    return int_to_bytes(data_length, 2) + data


async def decode_packet_id(reader) -> int:
    packet_id_bytes = await read_or_raise(reader, 2)
    packet_id = unpack("!H", packet_id_bytes)
    return packet_id[0]


def int_to_bytes_str(value: int) -> bytes:
    return str(value).encode('utf-8')



import inspect
import typing

from joatmon.orm.constraint import Constraint
from joatmon.orm.field import Field
from joatmon.orm.index import Index
from joatmon.utility import get_converter


class Meta(type):
    __collection__ = 'meta'

    structured = True
    force = True

    qb = None

    def __new__(mcs, name, bases, dct):
        return super().__new__(mcs, name, bases, dct)

    def fields(cls, predicate=lambda x: True) -> typing.Dict[str, Field]:
        return {k: v for k, v in inspect.getmembers(cls, lambda x: isinstance(x, Field)) if predicate(v)}

    def constraints(cls, predicate=lambda x: True) -> typing.Dict[str, Constraint]:
        return {k: v for k, v in inspect.getmembers(cls, lambda x: isinstance(x, Constraint)) if predicate(v)}

    def indexes(cls, predicate=lambda x: True) -> typing.Dict[str, Index]:
        return {k: v for k, v in inspect.getmembers(cls, lambda x: isinstance(x, Index)) if predicate(v)}

    def query(cls):
        return cls.qb


def normalize_kwargs(meta, **kwargs):
    ret = {}

    fields = meta.fields(meta)
    for key in kwargs.keys():
        field = list(filter(lambda x: x[0] == key, fields.items()))
        if len(field) != 1:
            raise ValueError(f'field {key} has to be only one on the document')
        field = field[0][1]

        ret[key] = get_converter(field.dtype)(kwargs[key])
    return ret



import numpy as np

from joatmon.nn.core import Loss

__all__ = ['CCELoss']


class CCELoss(Loss):
    def __init__(self):
        super(CCELoss, self).__init__()

        self._loss = None

    def forward(self, prediction, target) -> np.array:
        self._loss = -(target * np.log(prediction.data) + (1 - target) * np.log(1 - prediction.data)).summation()
        return self._loss



from joatmon.plugin.core import Plugin


class Auth(Plugin):
    async def authenticate(self, issuer, audience, expire_at):
        raise NotImplementedError

    async def authorize(self, token, issuer, audience):
        raise NotImplementedError



import random

import cv2
import numpy as np

__all__ = ['Puzzle2048']

from joatmon.game.core import CoreEnv


class Puzzle2048(CoreEnv):
    def __init__(self, size):
        super().__init__()

        self.size = size
        self.matrix = [[0 for _ in range(self.size)] for _ in range(self.size)]
        self.reset()

    def __enter__(self, *args, **kwargs):
        pass

    def __exit__(self, *args, **kwargs):
        pass

    def __repr__(self):
        return str(self)

    def __str__(self):
        ret = '-' * 33 + '\n'
        for row in self.matrix:
            ret += '|'
            for val in row:
                if val != 0:
                    v = 2 ** val
                    ret += '{:7d}'.format(v)
                else:
                    ret += ' ' * 7
                ret += '|'
            ret += '\n'
            ret += '-' * 33 + '\n'
        # ret += '-' * 33
        return ret

    def _check(self):
        for row in range(self.size):
            for col in range(self.size):
                val = self.matrix[row][col]
                if val == 0:
                    return True

                if row == 0:
                    if col == 0:
                        vals = [self.matrix[row + 1][col], self.matrix[row][col + 1]]
                    elif col == self.size - 1:
                        vals = [self.matrix[row + 1][col], self.matrix[row][col - 1]]
                    else:
                        vals = [self.matrix[row + 1][col], self.matrix[row][col - 1], self.matrix[row][col + 1]]
                elif row == self.size - 1:
                    if col == 0:
                        vals = [self.matrix[row - 1][col], self.matrix[row][col + 1]]
                    elif col == self.size - 1:
                        vals = [self.matrix[row - 1][col], self.matrix[row][col - 1]]
                    else:
                        vals = [self.matrix[row - 1][col], self.matrix[row][col - 1], self.matrix[row][col + 1]]
                else:
                    if col == 0:
                        vals = [self.matrix[row - 1][col], self.matrix[row + 1][col], self.matrix[row][col + 1]]
                    elif col == self.size - 1:
                        vals = [self.matrix[row - 1][col], self.matrix[row + 1][col], self.matrix[row][col - 1]]
                    else:
                        vals = [self.matrix[row - 1][col], self.matrix[row + 1][col], self.matrix[row][col - 1], self.matrix[row][col + 1]]
                if val in vals:
                    return True
        return False

    def _fill(self, action):
        ret = False
        if action == 0:
            for row in range(self.size):
                for col in range(self.size - 1, 0, -1):
                    if self.matrix[row][col] == 0 and self.matrix[row][col - 1] != 0:
                        self.matrix[row][col] = self.matrix[row][col - 1]
                        self.matrix[row][col - 1] = 0
                        ret = True
        elif action == 1:
            for col in range(self.size):
                for row in range(self.size - 1, 0, -1):
                    if self.matrix[row][col] == 0 and self.matrix[row - 1][col] != 0:
                        self.matrix[row][col] = self.matrix[row - 1][col]
                        self.matrix[row - 1][col] = 0
                        ret = True
        elif action == 2:
            for row in range(self.size):
                for col in range(self.size - 1):
                    if self.matrix[row][col] == 0 and self.matrix[row][col + 1] != 0:
                        self.matrix[row][col] = self.matrix[row][col + 1]
                        self.matrix[row][col + 1] = 0
                        ret = True
        elif action == 3:
            for col in range(self.size):
                for row in range(self.size - 1):
                    if self.matrix[row][col] == 0 and self.matrix[row + 1][col] != 0:
                        self.matrix[row][col] = self.matrix[row + 1][col]
                        self.matrix[row + 1][col] = 0
                        ret = True
        else:
            pass
        return ret

    def _merge(self, action):
        reward = 9e-1
        if action == 0:
            for row in range(self.size):
                for col in range(self.size - 1, 0, -1):
                    val1 = self.matrix[row][col]
                    val2 = self.matrix[row][col - 1]
                    if val1 == val2 and val1 != 0 and val2 != 0:
                        self.matrix[row][col] += 1
                        self.matrix[row][col - 1] = 0
                        reward += 2 ** val1
        elif action == 1:
            for col in range(self.size):
                for row in range(self.size - 1, 0, -1):
                    val1 = self.matrix[row][col]
                    val2 = self.matrix[row - 1][col]
                    if val1 == val2 and val1 != 0 and val2 != 0:
                        self.matrix[row][col] += 1
                        self.matrix[row - 1][col] = 0
                        reward += 2 ** val1
        elif action == 2:
            for row in range(self.size):
                for col in range(self.size - 1):
                    val1 = self.matrix[row][col]
                    val2 = self.matrix[row][col + 1]
                    if val1 == val2 and val1 != 0 and val2 != 0:
                        self.matrix[row][col] += 1
                        self.matrix[row][col + 1] = 0
                        reward += 2 ** val1
        elif action == 3:
            for col in range(self.size):
                for row in range(self.size - 1):
                    val1 = self.matrix[row][col]
                    val2 = self.matrix[row + 1][col]
                    if val1 == val2 and val1 != 0 and val2 != 0:
                        self.matrix[row][col] += 1
                        self.matrix[row + 1][col] = 0
                        reward += 2 ** val1
        else:
            pass
        return reward

    def _randomize(self):
        indexes = []
        for row in range(self.size):
            for col in range(self.size):
                if self.matrix[row][col] == 0:
                    indexes.append((row, col))
        if indexes:
            row, col = random.sample(indexes, 1)[0]
            self.matrix[row][col] = np.random.choice([1, 2], p=[0.9, 0.1])

    def close(self):
        pass

    def render(self, mode='human'):
        # print(self)
        rect_size = 75
        state_size = rect_size * self.size
        state = np.zeros((state_size, state_size, 3)).astype('uint8')

        for row in range(self.size):
            for col in range(self.size):
                cv2.rectangle(state, (col * rect_size, row * rect_size), ((col + 1) * rect_size, (row + 1) * rect_size), (255, 255, 255), 1)
                if self.matrix[row][col] != 0:
                    w, h = cv2.getTextSize(str(2 ** self.matrix[row][col]), cv2.FONT_HERSHEY_SIMPLEX, 0.5, 1)
                    cv2.putText(
                        state, str(2 ** self.matrix[row][col]), (int((col + 1) * rect_size - w[0]), int((row + 0.5) * rect_size + h)),
                        cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 255, 255), 1, cv2.LINE_AA
                    )
        cv2.imshow('state', state)
        cv2.waitKey(1)

    def reset(self):
        self.matrix = [[0 for _ in range(self.size)] for _ in range(self.size)]
        self._randomize()
        self._randomize()

        return self.matrix

    def seed(self, seed=None):
        pass

    def step(self, action):
        # 0 -> right
        # 1 -> down
        # 2 -> left
        # 3 -> up
        randomize = False
        # control fill, merge and randomize sometimes they do not work
        for _ in range(self.size):
            randomize |= self._fill(action)
        reward = self._merge(action)
        for _ in range(self.size):
            randomize |= self._fill(action)
        if randomize:
            self._randomize()
        return self.matrix, np.log(reward), not self._check(), {}



from joatmon.nn import functional as f

__all__ = ['Dropout']


class Dropout:
    def __init__(self, keep_prob):
        self._keep_prob = keep_prob

    def forward(self, inp):
        return f.dropout(
            inp=inp,
            keep_prob=self._keep_prob
        )



from joatmon.core import CoreException


class HZMQTTException(CoreException):
    pass


class MQTTException(CoreException):
    pass


class CodecException(CoreException):
    pass


class NoDataException(CoreException):
    pass



import json
import pickle

from joatmon.utility import (
    JSONEncoder,
    to_case,
    to_enumerable
)


class Serializable(object):
    def __init__(self, **kwargs):
        for slot_name, slot_value in kwargs.items():
            if not isinstance(slot_name, str):
                raise ValueError(f'{slot_name} is type {type(slot_name)} is not supported. only string type is supported for field names.')
            setattr(self, slot_name, slot_value)

    def __str__(self):
        return self.pretty_json

    def __repr__(self):
        return str(self)

    def keys(self):
        for key in self.__dict__.keys():
            yield key

    def values(self):
        for value in self.__dict__.values():
            yield value

    def items(self):
        for key, value in self.__dict__.items():
            yield key, value

    def __getitem__(self, key):
        return self.__dict__[key]

    @property
    def dict(self):
        return to_enumerable(self)

    @classmethod
    def from_dict(cls, dictionary: dict):
        return cls(**dictionary)

    @property
    def json(self) -> str:
        return json.dumps(self.dict, cls=JSONEncoder)

    @property
    def pretty_json(self) -> str:
        return json.dumps(self.dict, cls=JSONEncoder, indent=4)

    @classmethod
    def from_json(cls, string: str):
        return cls.from_dict(json.loads(string))

    @property
    def bytes(self) -> bytes:
        return pickle.dumps(self)

    @classmethod
    def from_bytes(cls, value: bytes):
        if value is None:
            return None

        obj = pickle.loads(value)
        return obj

    @property
    def snake(self):
        return to_case('snake', self.__dict__)

    @property
    def pascal(self):
        return to_case('pascal', self.__dict__)



from __future__ import print_function

from joatmon.assistant.task import BaseTask


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "ls",
            "description": "a function for user to list the given directory, if path is not given it will list the current working directory",
            "parameters": {
                "type": "object",
                "properties": {
                    "path": {
                        "type": "string",
                        "description": "directory to list"
                    }
                },
                "required": []
            }
        }

    def run(self):
        parent_os_path = self.kwargs.get('parent_os_path', '')
        os_path = self.kwargs.get('os_path', '')

        self.api.output(os_path)

        if not self.stop_event.is_set():
            self.stop_event.set()



arr_backend = None
nn_backend = None


def set_arr_backend(b):
    global arr_backend
    arr_backend = b
    return arr_backend


def set_nn_backend(b):
    global nn_backend
    nn_backend = b
    return nn_backend


def get_arr_backend():
    return arr_backend


def get_nn_backend():
    return nn_backend



from __future__ import print_function

from joatmon.assistant.service import BaseService
from joatmon.assistant.task import BaseTask


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "todo",
            "description": "a function for user to create a todo event and list them afterwards",
            "parameters": {
                "type": "object",
                "properties": {
                    "mode": {
                        "type": "string",
                        "enum": ["create", "list"]
                    }
                },
                "required": ["mode"]
            }
        }

    def run(self):
        mode = self.kwargs.get('mode', None)

        if mode == 'list':
            print(self.kwargs.get('mode', None), self.kwargs.get('event', None), self.kwargs.get('date', None))
        if mode == 'create':
            print(self.kwargs.get('mode', None), self.kwargs.get('event', None), self.kwargs.get('date', None))


class Service(BaseService):
    def __init__(self, api, **kwargs):
        super(Service, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {}

    def run(self):
        ...



import ctypes
import sys

import cv2
import numpy

__all__ = ['resolution', 'cursor', 'grab']

if sys.platform != 'win32':
    raise Exception('The screen module should only be used on a Windows system.')


class POINT(ctypes.Structure):
    _fields_ = [
        ("x", ctypes.c_long),
        ("y", ctypes.c_long)
    ]


def resolution():
    return ctypes.windll.user32.GetSystemMetrics(1), ctypes.windll.user32.GetSystemMetrics(0)


def cursor():
    _cursor = POINT()
    ctypes.windll.user32.GetCursorPos(ctypes.byref(_cursor))
    return _cursor.y, _cursor.x


def grab(region=None) -> numpy.ndarray:
    if sys.platform == 'win32':
        import win32api
        import win32con
        import win32gui
        import win32ui

        window_handle = win32gui.GetDesktopWindow()

        if region:
            left, top, right, bot = region
            width = right - left + 1
            height = bot - top + 1
        else:
            # width, height = resolution()
            # left = 0
            # top = 0

            width = win32api.GetSystemMetrics(win32con.SM_CXVIRTUALSCREEN)
            height = win32api.GetSystemMetrics(win32con.SM_CYVIRTUALSCREEN)
            left = win32api.GetSystemMetrics(win32con.SM_XVIRTUALSCREEN)
            top = win32api.GetSystemMetrics(win32con.SM_YVIRTUALSCREEN)

        window_handle_dc = win32gui.GetWindowDC(window_handle)
        source_dc = win32ui.CreateDCFromHandle(window_handle_dc)
        memory_dc = source_dc.CreateCompatibleDC()
        bit_map = win32ui.CreateBitmap()
        bit_map.CreateCompatibleBitmap(source_dc, width, height)
        memory_dc.SelectObject(bit_map)
        memory_dc.BitBlt((0, 0), (width, height), source_dc, (left, top), win32con.SRCCOPY)

        signed_ints_array = bit_map.GetBitmapBits(True)
        image = numpy.fromstring(signed_ints_array, dtype=numpy.uint8)
        image.shape = (height, width, 4)

        source_dc.DeleteDC()
        memory_dc.DeleteDC()
        win32gui.ReleaseDC(window_handle, window_handle_dc)
        win32gui.DeleteObject(bit_map.GetHandle())

        return cv2.cvtColor(image, cv2.COLOR_BGRA2BGR)



import functools


class Event:
    def __init__(self):
        self._subscribers = {}

    def __iadd__(self, fn):
        func = fn
        runnable = fn

        if isinstance(fn, functools.partial):
            func = fn.func

        if func.__qualname__ not in self._subscribers:
            self._subscribers[func.__qualname__] = runnable
        else:
            # might want to rethink this
            raise ValueError(f'{func.__qualname__} already subscribed to this event')

        return self

    def __isub__(self, fn):
        if isinstance(fn, functools.partial):
            fn = fn.func

        if fn.__qualname__ in self._subscribers:
            del self._subscribers[fn.__qualname__]
        else:
            # might want to rethink this
            raise ValueError(f'{fn.__qualname__} already unsubscribed to this event')

        return self

    def fire(self, *args, **kwargs):
        for name, fn in self._subscribers.items():
            fn(*args, **kwargs)


class AsyncEvent:
    def __init__(self):
        self._subscribers = {}

    def __iadd__(self, fn):
        if fn.__qualname__ not in self._subscribers:
            self._subscribers[fn.__qualname__] = fn
        else:
            # might want to rethink this
            raise ValueError(f'{fn.__qualname__} already subscribed to this event')

        return self

    def __isub__(self, fn):
        if fn.__qualname__ in self._subscribers:
            del self._subscribers[fn.__qualname__]
        else:
            # might want to rethink this
            raise ValueError(f'{fn.__qualname__} already unsubscribed to this event')

        return self

    async def fire(self, *args, **kwargs):
        for name, fn in self._subscribers.items():
            await fn(*args, **kwargs)



import torch
import torch.nn as nn
import torch.optim as optim

from joatmon.ai.models.core import CoreModel
from joatmon.ai.network import UNet
from joatmon.ai.utility import (
    load,
    save
)

__all__ = ['UNetModel']


class UNetModel(CoreModel):
    """
    U Network

    # Arguments
        models (`keras.nn.Model` instance): See [Model](#) for details.
        optimizer (`keras.optimizers.Optimizer` instance):
        See [Optimizer](#) for details.
        tau (float): tau.
        gamma (float): gamma.
    """

    def __init__(self, lr=1e-3, channels=3, classes=10):
        super().__init__()

        self.net = UNet(channels, classes, False)

        self.lr = lr

        self.optimizer = optim.Adam(self.net.parameters(), lr=self.lr)
        self.loss = nn.SmoothL1Loss()

    def load(self, path=''):
        load(self.net, path)

    def save(self, path=''):
        save(self.net, path)

    def initialize(self, w_init=None, b_init=None):
        for module in self.net.modules():
            if isinstance(module, (torch.nn.Conv2d, torch.nn.BatchNorm2d, torch.nn.Linear)):  # batch norm will be different
                if w_init:
                    torch.nn.init.kaiming_normal(module.weight)
                if b_init:  # bias init will be different
                    torch.nn.init.kaiming_normal(module.bias)

    def predict(self, state=None):
        ...

    def train(self, batch=None, update_target=False):
        ...

    def evaluate(self):
        ...



import pygame

from joatmon.event import Event


class InputField:
    def __init__(self):
        self._x = None
        self._y = None

        self._width = None
        self._height = None

        self._text = None
        self._font = None

        self._text_color = None
        self._background_color = None
        self._hover_color = None
        self._click_color = None

        self._border = None
        self._rect = None
        self._icon = None

        self._events = {
            'hover': Event(),
            'change': Event()
        }

        self.is_hovered = False
        self.is_clicked = False
        self.is_active = False

        self._placeholder = ''
        self._text = ''

    def position(self, x, y):
        self._x = x
        self._y = y

        return self

    def size(self, width, height):
        self._width = width
        self._height = height

        self._rect = pygame.Rect(self._x, self._y, self._width, self._height)

        return self

    def icon(self, icon):
        _icon = pygame.image.load(icon)

        self._width = self._height
        # self._rect = None
        self._rect = pygame.Rect(self._x, self._y, self._width, self._height)

        self._icon = pygame.transform.scale(_icon, (self._width - 10, self._height - 10))

        return self

    def text(self, text, font):
        self._text = text
        self._font = font

        return self

    def color(self, t_color, b_color, h_color, c_color):
        self._text_color = t_color
        self._background_color = b_color
        self._hover_color = h_color
        self._click_color = c_color

        return self

    def border(self, border):
        self._border = border
        return self

    def on(self, event, func):
        _event = self._events.get(event, None)
        if _event:
            _event += func

        return self

    def placeholder(self, placeholder):
        self._placeholder = placeholder

    def handle_event(self, event):
        if event.type == pygame.MOUSEMOTION:
            self.is_hovered = self._rect.collidepoint(event.pos)
            if self.is_hovered:
                _event = self._events.get('hover', None)
                if _event:
                    _event.fire()

        if event.type == pygame.MOUSEBUTTONDOWN:
            if self._rect.collidepoint(event.pos):
                self.is_active = True
            else:
                self.is_active = False
        elif event.type == pygame.KEYDOWN:
            if self.is_active:
                prev_text = self._text

                if event.key == pygame.K_RETURN:
                    self._text = ''
                elif event.key == pygame.K_BACKSPACE:
                    self._text = self._text[:-1]
                else:
                    self._text += event.unicode

                if prev_text != self._text:
                    _event = self._events.get('change', None)
                    if _event:
                        _event.fire()

    def draw(self, screen):
        border_color = self._border.get('color', (0, 0, 0))
        thickness = self._border.get('thickness', 1)
        radius = self._border.get('radius', 1)
        shadow = self._border.get('shadow', 3)

        if self._border:
            border_rect = self._rect.copy()
            border_rect.x -= thickness
            border_rect.y -= thickness
            border_rect.width += 2 * thickness
            border_rect.height += 2 * thickness

            if shadow:
                shadow_rect = border_rect.copy()
                shadow_rect.x += shadow
                shadow_rect.y += shadow
                pygame.draw.rect(screen, pygame.Color('gray'), shadow_rect, border_radius=radius)  # shadow color

                pygame.draw.rect(screen, pygame.Color('gray') if self.is_hovered else pygame.Color('white'), border_rect, border_radius=radius)
            pygame.draw.rect(screen, border_color, border_rect, thickness, border_radius=radius)

        font = pygame.font.SysFont(None, 24)

        # pygame.draw.rect(screen, self._background_color, self._rect, 0)
        # pygame.draw.rect(screen, self._border.get('color', (0, 0, 0)), self._rect, 2)

        text_surface = font.render(self._text if self._text is not None and self._text != '' else self._placeholder, True, self._text_color)
        screen.blit(text_surface, (self._rect.x + 5, self._rect.y + 5))



import random


class CoreBuffer(list):
    def __init__(self, values, batch_size):
        super(CoreBuffer, self).__init__()

        self.values = values
        self.batch_size = batch_size

    def __contains__(self, element):
        return element in self.values

    def __getitem__(self, idx):
        return self.values[idx]

    def __iter__(self):
        for value in self.values:
            yield value

    def __len__(self):
        return len(self.values)

    def add(self, element):
        self.values.append(element)

    def sample(self):
        return random.sample(self, self.batch_size)  # need to implement own random sampling algorithm


class CoreMemory(object):
    """
    Abstract base class for all implemented memory.

    Do not use this abstract base class directly but instead use one of the concrete memory implemented.

    To implement your own memory, you have to implement the following methods:

    - `remember`
    - `sample`
    """

    def __init__(self, buffer, batch_size):
        super(CoreMemory, self).__init__()

        if not isinstance(buffer, CoreBuffer):
            buffer = CoreBuffer([], batch_size)
        else:
            buffer.batch_size = batch_size

        self.buffer = buffer

    def __contains__(self, element):
        return element in self.buffer

    def __getitem__(self, idx):
        return self.buffer[idx]

    def __iter__(self):
        for value in self.buffer:
            yield value

    def __len__(self):
        return len(self.buffer)

    def remember(self, element):
        """
        Remember the transaction.

        Accepts a state, action, reward, next_state, terminal transaction.

        # Arguments
            transaction (abstract): state, action, reward, next_state, terminal transaction.
        """
        self.buffer.add(element)

    def sample(self):
        """
        Sample an experience replay batch with size.

        # Returns
            batch (abstract): Randomly selected batch
            from experience replay memory.
        """
        return self.buffer.sample()


class RingBuffer(CoreBuffer):
    def __init__(self, size, batch_size):
        self.data = [None] * (size + 1)
        self.start = 0
        self.end = 0

        super(RingBuffer, self).__init__(self.data, batch_size)

    def __contains__(self, item):
        # or return item in self.values
        for value in self:
            if value == item:
                return True

        return False

    def __getitem__(self, idx):
        if idx < 0 or idx >= len(self):
            raise KeyError

        idx = (self.start + idx) % len(self.data)
        return self.data[idx]

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __len__(self):
        if self.end < self.start:
            return self.end - self.start + len(self.data)
        else:
            return self.end - self.start

    def add(self, element):
        self.data[self.end] = element
        self.end = (self.end + 1) % len(self.data)

        if self.end == self.start:
            self.start = (self.start + 1) % len(self.data)


# should have option to, use PER and HER
class RingMemory(CoreMemory):
    """
    Ring Memory

    # Arguments
        size (int): .
    """

    def __init__(self, batch_size=32, size=960000):
        self.buffer = RingBuffer(size=size, batch_size=batch_size)

        super(RingMemory, self).__init__(self.buffer, batch_size)



class ConsoleReader:
    def __init__(self):
        super(ConsoleReader, self).__init__()

    def read(self):
        return input()


class ConsoleWriter:
    def __init__(self):
        super(ConsoleWriter, self).__init__()

    def write(self, data):
        print(data)
